from dlt.common.configuration import configspec
from dlt.common.destination import DestinationClientConfiguration, TLoaderFileFormat


@configspec(init=True)
class DummyClientConfiguration(DestinationClientConfiguration):
    destination_name: str = "dummy"
    loader_file_format: TLoaderFileFormat = "jsonl"
    fail_prob: float = 0.0
    retry_prob: float = 0.0
    completed_prob: float = 0.0
    timeout: float = 10.0



import os
from tempfile import mkdtemp

import dlt
from dlt.common.runners.venv import Venv

from examples.sources.singer_tap import tap

# create Venv with desired dependencies, in this case csv tap
# venv creation costs time so it should be created only once and reused

# here we use context manager to automatically delete venv after example was run
# the dependency is meltano version of csv tap
print("Spawning virtual environment to run singer and installing csv tap from git+https://github.com/MeltanoLabs/tap-csv.git")
# WARNING: on MACOS you need to have working gcc to use tap-csv, otherwise dependency will not be installed
with Venv.create(mkdtemp(), ["git+https://github.com/MeltanoLabs/tap-csv.git"]) as venv:
    # prep singer config for tap-csv
    csv_tap_config = {
        "files": [
            {
                "entity": "annotations_202205",
                "path": os.path.abspath("examples/data/singer_taps/model_annotations.csv"),
                "keys": [
                    "message id"
                ]
            }
        ]
    }
    print("running tap-csv")
    tap_source = tap(venv, "tap-csv", csv_tap_config, "examples/data/singer_taps/csv_catalog.json")
    info = dlt.pipeline("meltano_csv", destination="postgres").run(tap_source, credentials="postgres://loader@localhost:5432/dlt_data")
    print(info)



import os
import logging
from typing import Any, List, Sequence, Optional

from dlt.common import json
from dlt.common.typing import StrAny
from dlt.dbt_runner.exceptions import DBTRunnerException

# block disabling root logger
import logbook.compat
logbook.compat.redirect_logging = lambda : None

# can only import DBT after redirect is disabled
import dbt.main
import dbt.logger
from dbt.events import functions
from dbt.contracts import results as dbt_results
from dbt.exceptions import FailFastException


# keep this exception definition here due to mock of logbook
class DBTProcessingError(DBTRunnerException):
    def __init__(self, command: str, results: Any) -> None:
        self.command = command
        # the results from DBT may be anything
        self.results = results
        super().__init__(f"DBT command {command} could not be executed")


def initialize_dbt_logging(level: str, is_json_logging: bool) -> Sequence[str]:
    int_level = logging._nameToLevel[level]

    # wrap log setup to force out log level

    def setup_event_logger_wrapper(log_path: str, level_override:str = None) -> None:
        functions.setup_event_logger(log_path, level)
        # force log level as file is debug only
        functions.this.FILE_LOG.setLevel(level)
        functions.this.FILE_LOG.handlers[0].setLevel(level)

    dbt.main.setup_event_logger = setup_event_logger_wrapper

    globs = []
    if int_level <= logging.DEBUG:
        globs = ["--debug"]

    # return global parameters to be passed to setup logging

    if is_json_logging:
        return ["--log-format", "json"] + globs
    else:
        return globs


def is_incremental_schema_out_of_sync_error(error: dbt_results.RunResult) -> bool:
    return issubclass(type(error), dbt_results.RunResult) and error.status == dbt_results.RunStatus.Error and\
        "The source and target schemas on this incremental model are out of sync" in error.message


def run_dbt_command(package_path: str, command: str, profiles_dir: str, profile_name: Optional[str] = None,
                    global_args: Sequence[str] = None, command_args: Sequence[str] = None, dbt_vars: StrAny = None) -> Sequence[dbt_results.BaseResult]:
    args = ["--profiles-dir", profiles_dir]
    # add profile name if provided
    if profile_name:
        args += ["--profile", profile_name]
    # serialize dbt variables to pass to package
    if dbt_vars:
        args += ["--vars", json.dumps(dbt_vars)]
    if command_args:
        args += command_args

    # cwd to package dir
    working_dir = os.getcwd()
    os.chdir(package_path)
    try:
        results: List[dbt_results.BaseResult] = None
        success: bool = None
        results, success = dbt.main.handle_and_check((global_args or []) + [command] + args)  # type: ignore
        assert type(success) is bool
        if not success:
            raise DBTProcessingError(command ,results)
        return results
    except FailFastException as ff:
        raise DBTProcessingError(command, ff.result) from ff
    finally:
        # unblock logger manager to run next command
        dbt.logger.log_manager.reset_handlers()
        # go back to working dir
        os.chdir(working_dir)



from contextlib import contextmanager
from typing import Any, AnyStr, Iterator, List, Optional, Sequence
from dlt.common.typing import StrAny
from dlt.destinations.exceptions import DatabaseTerminalException, DatabaseTransientException, DatabaseUndefinedRelation
import google.cloud.bigquery as bigquery  # noqa: I250
from google.cloud.bigquery.dbapi import Connection as DbApiConnection
from google.cloud import exceptions as gcp_exceptions
from google.cloud.bigquery.dbapi import exceptions as dbapi_exceptions
from google.api_core import exceptions as api_core_exceptions

from dlt.common import logger
from dlt.common.configuration.specs import GcpClientCredentials

from dlt.destinations.typing import DBCursor
from dlt.destinations.sql_client import SqlClientBase, raise_database_error

# terminal reasons as returned in BQ gRPC error response
# https://cloud.google.com/bigquery/docs/error-messages
BQ_TERMINAL_REASONS = ["billingTierLimitExceeded", "duplicate", "invalid", "notFound", "notImplemented", "stopped", "tableUnavailable"]
# invalidQuery is an transient error -> must be fixed by programmer


class BigQuerySqlClient(SqlClientBase[bigquery.Client]):
    def __init__(self, default_dataset_name: str, credentials: GcpClientCredentials) -> None:
        self._client: bigquery.Client = None
        self.credentials: GcpClientCredentials = credentials
        super().__init__(default_dataset_name)

        self.default_retry = bigquery.DEFAULT_RETRY.with_deadline(credentials.retry_deadline)
        self.default_query = bigquery.QueryJobConfig(default_dataset=self.fully_qualified_dataset_name())

    def open_connection(self) -> None:
        self._client = bigquery.Client(
            self.credentials.project_id,
            credentials=self.credentials.to_service_account_credentials(),
            location=self.credentials.location
        )

        # patch the client query so our defaults are used
        query_orig = self._client.query

        def query_patch(
            query: str,
            retry: Any = self.default_retry,
            timeout: Any = self.credentials.http_timeout,
            **kwargs: Any
        ) -> Any:
            return query_orig(query, retry=retry, timeout=timeout, **kwargs)

        self._client.query = query_patch


    def close_connection(self) -> None:
        if self._client:
            self._client.close()
            self._client = None

    @property
    def native_connection(self) -> bigquery.Client:
        return self._client

    def has_dataset(self) -> bool:
        try:
            self._client.get_dataset(self.fully_qualified_dataset_name(), retry=self.default_retry, timeout=self.credentials.http_timeout)
            return True
        except gcp_exceptions.NotFound:
            return False

    def create_dataset(self) -> None:
        self._client.create_dataset(
            self.fully_qualified_dataset_name(),
            exists_ok=False,
            retry=self.default_retry,
            timeout=self.credentials.http_timeout
        )

    def drop_dataset(self) -> None:
        self._client.delete_dataset(
            self.fully_qualified_dataset_name(),
            not_found_ok=True,
            delete_contents=True,
            retry=self.default_retry,
            timeout=self.credentials.http_timeout
        )

    # @raise_database_error
    def execute_sql(self, sql: AnyStr, *args: Any, **kwargs: Any) -> Optional[Sequence[Sequence[Any]]]:
        with self.execute_query(sql, *args, **kwargs) as curr:
            if not curr.description:
                return None
            else:
                try:
                    f = curr.fetchall()
                    return f
                except api_core_exceptions.InvalidArgument as ia_ex:
                    if "non-table entities cannot be read" in str(ia_ex):
                        return None
                    raise

    @contextmanager
    @raise_database_error
    def execute_query(self, query: AnyStr,  *args: Any, **kwargs: Any) -> Iterator[DBCursor]:
        conn: DbApiConnection = None
        curr: DBCursor = None
        db_args = args if args else kwargs if kwargs else None
        logger.debug(f"Will execute query {query} {db_args}")  # type: ignore
        try:
            conn = DbApiConnection(client=self._client)
            curr = conn.cursor()
            curr.execute(query, db_args, job_config=self.default_query)
            yield curr
        finally:
            if conn:
                # will also close all cursors
                conn.close()

    def fully_qualified_dataset_name(self) -> str:
        return f"{self.credentials.project_id}.{self.default_dataset_name}"

    @classmethod
    def _make_database_exception(cls, ex: Exception) -> Exception:
        if cls.is_dbapi_exception(ex):
            # google cloud exception in first argument: https://github.com/googleapis/python-bigquery/blob/main/google/cloud/bigquery/dbapi/cursor.py#L205
            cloud_ex = ex.args[0]
            reason = cls._get_reason_from_errors(cloud_ex)
            if reason is None:
                if isinstance(ex, (dbapi_exceptions.DataError, dbapi_exceptions.IntegrityError)):
                    return DatabaseTerminalException(ex)
                elif isinstance(ex, dbapi_exceptions.ProgrammingError):
                    return DatabaseTransientException(ex)
            if reason == "notFound":
                return DatabaseUndefinedRelation(ex)
            if reason == "invalidQuery" and "Unrecognized name" in str(ex):
                # unknown column etc.
                return DatabaseTerminalException(ex)
            elif reason in BQ_TERMINAL_REASONS:
                return DatabaseTerminalException(ex)
            # anything else is transient
            return DatabaseTransientException(ex)
        return ex

    @staticmethod
    def _get_reason_from_errors(gace: api_core_exceptions.GoogleAPICallError) -> Optional[str]:
        errors: List[StrAny] = getattr(gace, "errors", None)
        if errors and isinstance(errors, Sequence):
            return errors[0].get("reason")  # type: ignore
        return None

    @staticmethod
    def is_dbapi_exception(ex: Exception) -> bool:
        return isinstance(ex, dbapi_exceptions.Error)



from subprocess import PIPE, CalledProcessError
from typing import Any, Iterator

from dlt.common.runners.venv import Venv


def iter_stdout(venv: Venv, command: str, *script_args: Any) -> Iterator[str]:
    # start a process in virtual environment, assume that text comes from stdout
    with venv.start_command(command, *script_args, stdout=PIPE, stderr=PIPE, text=True) as process:
        exit_code: int = None

        # read all the lines until empty marker is returned
        for line in iter(process.stdout.readline, ''):
            yield line[:-1]

        # we fail iterator if exit code is not 0
        exit_code = process.wait()
        if exit_code != 0:
            raise CalledProcessError(exit_code, command)



# # ─────── DLT ───────────────
# # ───▄▄█████████▄────────────
# # ──███████▄██▀▀─────────────
# # ─▐████████── JSON ── JSON──
# # ──█████████▄▄──────────────
# # ───▀██████████▀────────────

# from typing import Sequence

# from dlt.common.typing import DictStrAny
# from dlt.common import json
# from dlt.common.schema import Schema
# from dlt.pipeline import Pipeline, GCPPipelineCredentials

# # the load schema will be named {pipeline_name}_{source_name}
# # this allows you to easily consume multiple environments/instances of the same source
# schema_prefix = 'library_documents'
# schema_source_suffix = 'prod'

# # you authenticate by passing a credential, such as RDBMS host/user/port/pass or gcp service account json.
# gcp_credential_json_file_path = "/Users/adrian/PycharmProjects/sv/dlt/temp/scalevector-1235ac340b0b.json"

# # get some data and give a name to the parent table.
# # If you have a stream with different document types, pass the doc type into the table name to separate the stream.
# # In our example we have a list of 2 books

# data_file_path = "/Users/adrian/PycharmProjects/sv/dlt/examples/data/demo_example.json"
# parent_table = 'books'

# data_schema = None
# data_schema_file_path = "/Users/adrian/PycharmProjects/sv/dlt/examples/schemas/inferred_demo_schema.yml"

# # or pass a saved data schema file
# # f = open(data_schema_file_path, "r", encoding="utf-8")
# # data_schema = f.read()
# #
# #

# """
# to get your data in a flat table, you need to join the sub-json back to the parent
# example sql:

# SELECT
#     b.* EXCEPT(_dlt_load_id, _dlt_id, _dlt_root_id),
#     bc.* EXCEPT(_dlt_parent_id, _dlt_id, _dlt_root_id),
# FROM `scalevector.library_documents_prod.books` as b
# left join `scalevector.library_documents_prod.books__category` as bc
# on b._dlt_id = bc._dlt_parent_id

# SQL result rows:
# {  "isbn": "123-456-222",  "author__lastname": "Panda",  "author__firstname": "Jane",  "editor__lastname": "Smite",  "editor__firstname": "Jane",  "title": "The Ultimate Database Study Guide",  "value": "Non-Fiction",  "_dlt_list_idx": "0"}
# {  "isbn": "123-456-222",  "author__lastname": "Panda",  "author__firstname": "Jane",  "editor__lastname": "Smite",  "editor__firstname": "Jane",  "title": "The Ultimate Database Study Guide",  "value": "Technology",  "_dlt_list_idx": "1"}
# {  "isbn": "123-456-789",  "author__lastname": "Jayson",  "author__firstname": "Joe",  "editor__lastname": "Smite",  "editor__firstname": "Jane",  "title": "Json for big data",  "value": "SF",  "_dlt_list_idx": "0"}
# {  "isbn": "123-456-789",  "author__lastname": "Jayson",  "author__firstname": "Joe",  "editor__lastname": "Smite",  "editor__firstname": "Jane",  "title": "Json for big data",  "value": "Horror",  "_dlt_list_idx": "1"}
# {  "isbn": "123-456-789",  "author__lastname": "Jayson",  "author__firstname": "Joe",  "editor__lastname": "Smite",  "editor__firstname": "Jane",  "title": "Json for big data",  "value": "Dystopia",  "_dlt_list_idx": "2"}
# """  # noqa: E501


# if __name__ == "__main__":

#     # loading and error handling below:

#     def get_json_file_data(path: str) -> Sequence[DictStrAny]:
#         with open(path, "r", encoding="utf-8") as f:
#             data = json.load(f)
#         return data  # type: ignore

#     data = get_json_file_data(data_file_path)

#     # this is example of extracting from iterator that may be a
#     # - jsonl file you read line by line
#     # - a list of dicts
#     # - iterator over panda frame https://pandas.pydata.org/docs/reference/api/pandas.DataFrame.to_dict.html
#     # - any iterator with map function
#     # - any iterator whatsoever ie getting data from REST API

#     # extracting from iterator always happens in single worker thread and iterator is directly returning data so there's no retry mechanism if anything fails

#     # working BQ creds
#     # credentials = Pipeline.load_gcp_credentials("_secrets/project1234_service.json", "gamma_guild")

#     # working redshift creds, you can pass password as last parameter or via PG__PASSWORD env variable ie.
#     # LOG_LEVEL=INFO PG__PASSWORD=.... python examples/discord_iterator.py
#     credentials = GCPPipelineCredentials.from_services_file(gcp_credential_json_file_path, schema_prefix)

#     pipeline = Pipeline(schema_source_suffix)

#     schema: Schema = None
#     # uncomment to use already modified schema, but the auto-inferred schema will also work nicely
#     # TODO: schema below needs is not yet finished, more exploration needed
#     # schema = Pipeline.load_schema_from_file("examples/schemas/discord_schema.yml")
#     pipeline.create_pipeline(credentials, schema=schema)
#     # the pipeline created a working directory. you can always attach the pipeline back by just providing the dir to
#     # Pipeline::restore_pipeline
#     print(pipeline.root_path)

#     # and extract it
#     pipeline.extract(iter(data), table_name=parent_table)

#     # now create loading packages and infer the schema
#     pipeline.normalize()

#     schema = pipeline.get_default_schema()
#     schema_yaml = schema.to_pretty_yaml()
#     f = open(data_schema_file_path, "a", encoding="utf-8")
#     f.write(schema_yaml)
#     f.close()
#     # pipeline.save_schema_to_file(data_schema_file_path, schema)

#     # show loads, each load contains a copy of the schema that corresponds to the data inside
#     # and a set of directories for job states (new -> in progress -> failed|completed)
#     new_loads = pipeline.list_normalized_loads()
#     print(new_loads)

#     # load packages
#     pipeline.load()

#     # should be empty
#     new_loads = pipeline.list_normalized_loads()
#     print(new_loads)

#     # now enumerate all complete loads if we have any failed packages
#     # complete but failed job will not raise any exceptions
#     completed_loads = pipeline.list_completed_loads()
#     # print(completed_loads)
#     for load_id in completed_loads:
#         print(f"Checking failed jobs in {load_id}")
#         for job, failed_message in pipeline.list_failed_jobs(load_id):
#             print(f"JOB: {job}\nMSG: {failed_message}")



import base64
import os

import dlt

"""
This is example of ad-hoc pipeline where the data is directly passed to `run` function, schema hints are explicit and credentials are passed directly
or with the use of environment variables
"""

# 1. configuration: name your dataset, table, pass credentials
dataset_name = 'dlt_quickstart'
pipeline_name = 'dlt_quickstart'
table_name = 'my_json_doc'

gcp_credentials_json = {
    "type": "service_account",
    "project_id": "zinc-mantra-353207",
    "private_key": "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",  # noqa
    "client_email": "data-load-tool-public-demo@zinc-mantra-353207.iam.gserviceaccount.com",
}

dsn = "postgres://loader:loader@localhost:5432/dlt_data"

# we do not want to have this key verbatim in repo so we decode it here
gcp_credentials_json["private_key"] = bytes([_a ^ _b for _a, _b in zip(base64.b64decode(gcp_credentials_json["private_key"]), b"quickstart-sv"*150)]).decode("utf-8")

# enable the automatic schema export so you can see the auto-generated schema
export_schema_path = "examples/schemas/"


# 2. Create a pipeline
pipeline = dlt.pipeline(pipeline_name, destination="postgres", dataset_name=dataset_name, export_schema_path=export_schema_path, full_refresh=True)


# 3. Pass the data to the pipeline and give it a table name. Optionally normalize and handle schema.

rows = [{"name": "Ana", "age": 30, "id": 456, "children": [{"name": "Bill", "id": 625},
                                                           {"name": "Elli", "id": 591}
                                                           ]},

        {"name": "Bob", "age": 30, "id": 455, "children": [{"name": "Bill", "id": 625},
                                                           {"name": "Dave", "id": 621}
                                                           ]}
        ]

load_info = pipeline.run(rows, credentials=dsn, table_name=table_name, write_disposition="replace")

# 4. Optional error handling - print, raise or handle.
print()
print(load_info)
print()


# show to auto-generated schema
# print(pipeline.default_schema.to_pretty_yaml())

# now you can use your data

# Tables created:
print("Now lets SELECT some data:")
print()
with pipeline.sql_client(credentials=dsn) as c:
    query = "SELECT * FROM my_json_doc"
    df = c.execute_sql(query)
    print(query)
    print(list(df))
    print()
    # {  "name": "Ana",  "age": "30",  "id": "456",  "_dlt_load_id": "1654787700.406905",  "_dlt_id": "5b018c1ba3364279a0ca1a231fbd8d90"}
    # {  "name": "Bob",  "age": "30",  "id": "455",  "_dlt_load_id": "1654787700.406905",  "_dlt_id": "afc8506472a14a529bf3e6ebba3e0a9e"}

    query = "SELECT * FROM my_json_doc__children LIMIT 1000"
    df = c.execute_sql(query)
    print(query)
    print(list(df))
    print()
    # {"name": "Bill", "id": "625", "_dlt_parent_id": "5b018c1ba3364279a0ca1a231fbd8d90", "_dlt_list_idx": "0", "_dlt_root_id": "5b018c1ba3364279a0ca1a231fbd8d90",
    #   "_dlt_id": "7993452627a98814cc7091f2c51faf5c"}
    # {"name": "Bill", "id": "625", "_dlt_parent_id": "afc8506472a14a529bf3e6ebba3e0a9e", "_dlt_list_idx": "0", "_dlt_root_id": "afc8506472a14a529bf3e6ebba3e0a9e",
    #   "_dlt_id": "9a2fd144227e70e3aa09467e2358f934"}
    # {"name": "Dave", "id": "621", "_dlt_parent_id": "afc8506472a14a529bf3e6ebba3e0a9e", "_dlt_list_idx": "1", "_dlt_root_id": "afc8506472a14a529bf3e6ebba3e0a9e",
    #   "_dlt_id": "28002ed6792470ea8caf2d6b6393b4f9"}
    # {"name": "Elli", "id": "591", "_dlt_parent_id": "5b018c1ba3364279a0ca1a231fbd8d90", "_dlt_list_idx": "1", "_dlt_root_id": "5b018c1ba3364279a0ca1a231fbd8d90",
    #   "_dlt_id": "d18172353fba1a492c739a7789a786cf"}

    # and we can join them via auto generated keys
    query = """
        select p.name, p.age, p.id as parent_id,
            c.name as child_name, c.id as child_id, c._dlt_list_idx as child_order_in_list
        from my_json_doc as p
        left join my_json_doc__children  as c
            on p._dlt_id = c._dlt_parent_id
    """
    df = c.execute_sql(query)
    print(query)
    print(list(df))
    print()
    # {  "name": "Ana",  "age": "30",  "parent_id": "456",  "child_name": "Bill",  "child_id": "625",  "child_order_in_list": "0"}
    # {  "name": "Ana",  "age": "30",  "parent_id": "456",  "child_name": "Elli",  "child_id": "591",  "child_order_in_list": "1"}
    # {  "name": "Bob",  "age": "30",  "parent_id": "455",  "child_name": "Bill",  "child_id": "625",  "child_order_in_list": "0"}
    # {  "name": "Bob",  "age": "30",  "parent_id": "455",  "child_name": "Dave",  "child_id": "621",  "child_order_in_list": "1"}



import os
import tempfile
from typing import Iterator, Optional
from git import Repo, RepositoryDirtyError
from contextlib import contextmanager

from dlt.common.utils import uniq_id


@contextmanager
def git_custom_key_command(private_key: Optional[str]) -> Iterator[str]:
    if private_key:
        key_file = tempfile.mktemp(prefix=uniq_id())
        with open(key_file, "w", encoding="utf-8") as f:
            f.write(private_key)
        try:
            # permissions so SSH does not complain
            os.chmod(key_file, 0o600)
            yield 'ssh -o "StrictHostKeyChecking accept-new" -i %s' % key_file
        finally:
            os.remove(key_file)
    else:
        yield 'ssh -o "StrictHostKeyChecking accept-new"'


def ensure_remote_head(repo_path: str, with_git_command: Optional[str] = None) -> None:
    # update remotes and check if heads are same. ignores locally modified files
    repo = Repo(repo_path)
    # use custom environment if specified
    with repo.git.custom_environment(GIT_SSH_COMMAND=with_git_command):
        # update origin
        repo.remote().update()
        # get branch status
        status: str = repo.git.status("--short", "--branch", "-uno")
        # we expect first status line ## main...origin/main
        status_line = status.split("/n")[0]
        if not (status_line.startswith("##") and not status_line.endswith("]")):
            raise RepositoryDirtyError(repo, status)


def clone_repo(repository_url: str, clone_path: str, branch: Optional[str] = None, with_git_command: Optional[str] = None) -> None:
    repo = Repo.clone_from(repository_url, clone_path, env=dict(GIT_SSH_COMMAND=with_git_command))
    if branch:
        repo.git.checkout(branch)



import inspect
import os
import ast
import click
import shutil
import tempfile
import sys
from types import ModuleType
from typing import Dict, List, Tuple
from importlib.metadata import version as pkg_version
from importlib import import_module

from dlt.common.git import clone_repo
from dlt.common.configuration.providers.toml import ConfigTomlProvider, SecretsTomlProvider
from dlt.common.configuration.specs.base_configuration import is_secret_hint
from dlt.common.configuration.accessors import DLT_SECRETS_VALUE, DLT_CONFIG_VALUE
from dlt.common.exceptions import DltException
from dlt.common.logger import DLT_PKG_NAME
from dlt.common.normalizers.names.snake_case import normalize_schema_name
from dlt.common.destination import DestinationReference
from dlt.common.reflection.utils import set_ast_parents
from dlt.common.schema.exceptions import InvalidSchemaName
from dlt.common.storages.file_storage import FileStorage
from dlt.common.typing import AnyType, is_optional_type

from dlt.extract.decorators import _SOURCES
import dlt.reflection.names as n
from dlt.reflection.script_inspector import inspect_pipeline_script
from dlt.reflection.script_visitor import PipelineScriptVisitor

import dlt.cli.echo as fmt
from dlt.cli.config_toml_writer import WritableConfigValue, write_values


REQUIREMENTS_TXT = "requirements.txt"
PYPROJECT_TOML = "pyproject.toml"


def _clone_init_repo(branch: str) -> Tuple[FileStorage, List[str], str]:
    # return tuple is (file storage for cloned repo, list of template files to copy, the default pipeline template script)
    # template_dir = "~/src/python-dlt-init-template"
    template_dir = tempfile.mkdtemp()
    clone_repo("https://github.com/scale-vector/python-dlt-init-template.git", template_dir, branch=branch)

    clone_storage = FileStorage(template_dir)

    assert clone_storage.has_file("pipeline.py")

    # import the settings from the clone
    try:
        template_dir, template_module_name = os.path.split(template_dir)
        sys.path.append(template_dir)
        template_module = import_module(template_module_name)
        return clone_storage, template_module.TEMPLATE_FILES, template_module.PIPELINE_SCRIPT
    finally:
        sys.path.remove(template_dir)


def _parse_init_script(script_source: str, init_script_name: str) -> PipelineScriptVisitor:
    # parse the script first
    tree = ast.parse(source=script_source)
    set_ast_parents(tree)
    visitor = PipelineScriptVisitor(script_source)
    visitor.visit(tree)
    if len(visitor.mod_aliases) == 0:
        raise CliCommandException("init", f"The pipeline script {init_script_name} does not import dlt or has bizarre import structure")
    if visitor.is_destination_imported:
        raise CliCommandException("init", f"The pipeline script {init_script_name} import a destination from dlt.destinations. You should specify destinations by name when calling dlt.pipeline or dlt.run in init scripts.")
    if n.PIPELINE not in visitor.known_calls:
        raise CliCommandException("init", f"The pipeline script {init_script_name} does not seem to initialize pipeline with dlt.pipeline. Please initialize pipeline explicitly in init scripts.")
    if n.RUN not in visitor.known_calls:
        raise CliCommandException("init", f"The pipeline script {init_script_name} does not seem to run the pipeline.")

    return visitor


def _find_argument_nodes_to_replace(visitor: PipelineScriptVisitor, replace_nodes: List[Tuple[str, str]], init_script_name: str) -> List[Tuple[ast.Constant, str, str]]:
    # the input tuple (call argument name, replacement value)
    # the returned tuple (node, replacement value, node type)
    transformed_nodes: List[Tuple[ast.Constant, str, str]] = []
    known_calls: Dict[str, List[inspect.BoundArguments]] = visitor.known_calls
    for arg_name, calls in known_calls.items():
        for args in calls:
            for t_arg_name, t_value in replace_nodes:
                if t_arg_name in args.arguments:
                    dn_node: ast.AST = args.arguments[t_arg_name]
                    if dn_node is not None:
                        if not isinstance(dn_node, ast.Constant) or not isinstance(dn_node.value, str):
                            raise CliCommandException("init", f"The pipeline script {init_script_name} must pass the {t_arg_name} as string to '{arg_name}' function in line {dn_node.lineno}")
                        else:
                            transformed_nodes.append((dn_node, t_value, t_arg_name))

    # there was at least one replacement
    for t_arg_name, _ in replace_nodes:
        if len(list(filter(lambda tn: tn[2] == t_arg_name, transformed_nodes))) == 0:
            raise CliCommandException("init", f"The pipeline script {init_script_name} is not explicitly passing the '{t_arg_name}' argument to 'pipeline' or 'run' function. In init script the default and configured values are not accepted.")
    return transformed_nodes


def _detect_required_configs(visitor: PipelineScriptVisitor, script_module: ModuleType, init_script_name: str) -> Tuple[Dict[str, WritableConfigValue], Dict[str, WritableConfigValue]]:
    # all detected secrets with namespaces
    required_secrets: Dict[str, WritableConfigValue] = {}
    # all detected configs with namespaces
    required_config: Dict[str, WritableConfigValue] = {}

    # skip sources without spec. those are not imported and most probably are inner functions. also skip the sources that are not called
    # also skip the sources that are called from functions, the parent of call object to the source must be None (no outer function)
    known_imported_sources = {name: _SOURCES[name] for name in visitor.known_sources
        if name in _SOURCES and name in visitor.known_source_calls and any(call.parent is None for call in visitor.known_source_calls[name])}  # type: ignore

    for source_name, source_info in known_imported_sources.items():
        source_config = source_info.SPEC()
        spec_fields = source_config.get_resolvable_fields()
        for field_name, field_type in spec_fields.items():
            val_store = None
            # all secrets must go to secrets.toml
            if is_secret_hint(field_type):
                val_store = required_secrets
            # all configs that are required and do not have a default value must go to config.toml
            elif not is_optional_type(field_type) and getattr(source_config, field_name) is None:
                val_store = required_config

            if val_store is not None:
                # we are sure that all resources come from single file so we can put them in single namespace
                # namespaces = () if len(known_imported_sources) == 1 else ("sources", source_name)
                val_store[source_name + ":" + field_name] = WritableConfigValue(field_name, field_type, ())

    return required_secrets, required_config


def _rewrite_script(script_source: str, transformed_nodes: List[Tuple[ast.Constant, str, str]]) -> str:
    module_source_lines: List[str] = ast._splitlines_no_ff(script_source)  # type: ignore
    script_lines: List[str] = []
    last_line = -1
    last_offset = -1
    # sort transformed nodes by line and offset
    for node, t_value, _ in sorted(transformed_nodes, key=lambda n: (n[0].lineno, n[0].col_offset)):
        # do we have a line changed
        if last_line != node.lineno - 1:
            # add remainder from the previous line
            if last_offset >= 0:
                script_lines.append(module_source_lines[last_line][last_offset:])
            # add all new lines from previous line to current
            script_lines.extend(module_source_lines[last_line+1:node.lineno-1])
            # add trailing characters until node in current line starts
            script_lines.append(module_source_lines[node.lineno-1][:node.col_offset])
        elif last_offset >= 0:
            # no line change, add the characters from the end of previous node to the current
            script_lines.append(module_source_lines[last_line][last_offset:node.col_offset])

        # replace node value
        script_lines.append(f'"{t_value}"')
        last_line = node.end_lineno - 1
        last_offset = node.end_col_offset

    # add all that was missing
    if last_offset >= 0:
        script_lines.append(module_source_lines[last_line][last_offset:])
    script_lines.extend(module_source_lines[last_line+1:])

    dest_script = "".join(script_lines)
    # validate by parsing
    ast.parse(source=dest_script)
    return dest_script


def init_command(pipeline_name: str, destination_name: str, branch: str) -> None:
    # try to import the destination and get config spec
    destination_reference = DestinationReference.from_name(destination_name)
    destination_spec = destination_reference.spec()

    click.echo("Cloning the init scripts...")
    clone_storage, TEMPLATE_FILES, PIPELINE_SCRIPT = _clone_init_repo(branch)

    # get init script variant or the default
    init_script_name = os.path.join("variants", pipeline_name + ".py")
    if clone_storage.has_file(init_script_name):
        # use variant
        dest_pipeline_script = pipeline_name + ".py"
        click.echo(f"Using a init script variant {fmt.bold(dest_pipeline_script)}")
    else:
        # use default
        init_script_name = PIPELINE_SCRIPT

    # normalize source name
    norm_source_name = normalize_schema_name(pipeline_name)
    if norm_source_name != pipeline_name:
        raise InvalidSchemaName(pipeline_name, norm_source_name)
    dest_pipeline_script = norm_source_name + ".py"

    # prepare destination storage
    dest_storage = FileStorage(os.path.abspath(os.path.join(".")))
    if not dest_storage.has_folder(".dlt"):
        dest_storage.create_folder(".dlt")

    # check if directory is empty
    toml_files = [".dlt/config.toml", ".dlt/secrets.toml"]
    created_files = TEMPLATE_FILES + [dest_pipeline_script] + toml_files
    existing_files = dest_storage.list_folder_files(".", to_root=False) + dest_storage.list_folder_files(".dlt", to_root=True)
    will_overwrite = set(created_files).intersection(existing_files)
    if will_overwrite:
        if not click.confirm(f"The following files in current folder will be replaced: {will_overwrite}. Do you want to continue?", default=False):
            raise FileExistsError("Would overwrite following files:", will_overwrite)

    # read module source and parse it
    visitor = _parse_init_script(clone_storage.load(init_script_name), init_script_name)

    # find all arguments in all calls to replace
    transformed_nodes = _find_argument_nodes_to_replace(
        visitor,
        [("destination", destination_name), ("pipeline_name", pipeline_name), ("dataset_name", pipeline_name + "_data")],
        init_script_name
    )

    # inspect the script
    script_module = inspect_pipeline_script(clone_storage.make_full_path(init_script_name))

    if len(_SOURCES) == 0:
        raise CliCommandException("init", f"The pipeline script {init_script_name} is not creating or importing any sources or resources")

    for source_q_name, source_config in _SOURCES.items():
        if source_q_name not in visitor.known_sources:
            print(visitor.known_sources)
            raise CliCommandException("init", f"The pipeline script {init_script_name} imports a source/resource {source_config.f.__name__} from module {source_config.module.__name__}. In init scripts you must declare all sources and resources in single file.")

    # detect all the required secrets and configs that should go into tomls files
    required_secrets, required_config = _detect_required_configs(visitor, script_module, init_script_name)
    # add destination spec to required secrets
    credentials_type = destination_spec().get_resolvable_fields()["credentials"]
    required_secrets["destinations:" + destination_name] = WritableConfigValue("credentials", credentials_type, ("destination", destination_name))

    # modify the script
    dest_script_source = _rewrite_script(visitor.source, transformed_nodes)

    # welcome message
    click.echo()
    click.echo("Your new pipeline %s is ready to be customized!" % fmt.bold(pipeline_name))
    click.echo("* Review and change how dlt loads your data in %s" % fmt.bold(dest_pipeline_script))
    click.echo("* Add credentials to %s and other secrets in %s" % (fmt.bold(destination_name), fmt.bold(toml_files[1])))
    click.echo("* Configure your pipeline in %s" % fmt.bold(toml_files[1]))
    click.echo("* See %s for further information" % fmt.bold("README.md"))
    click.echo()

    # add dlt to dependencies
    dlt_version = pkg_version(DLT_PKG_NAME)
    requirements_txt: str = None
    # figure out the build system
    if dest_storage.has_file(PYPROJECT_TOML):
       click.echo("Your python dependencies are kept in %s. Please add the dependency for %s as follows:" % (fmt.bold(PYPROJECT_TOML), fmt.bold(DLT_PKG_NAME)))
       click.echo(fmt.bold("%s [%s] >= %s" % (DLT_PKG_NAME, destination_name, dlt_version)))
       click.echo("If you are using poetry you may issue the following command:")
       click.echo(fmt.bold("poetry add %s -E %s" % (DLT_PKG_NAME, destination_name)))
       click.echo("If the dependency is already added, make sure you add the extra %s to it" % fmt.bold(destination_name))
    else:
        req_dep_line = f"{DLT_PKG_NAME}[{destination_name}] >= {pkg_version(DLT_PKG_NAME)}\n"
        if dest_storage.has_file(REQUIREMENTS_TXT):
            click.echo("Your python dependencies are kept in %s. Please add the dependency for %s as follows:" % (fmt.bold(REQUIREMENTS_TXT), fmt.bold(DLT_PKG_NAME)))
            click.echo(req_dep_line)
            click.echo("To install dlt with the %s extra using pip:" % fmt.bold(destination_name))
        else:
            requirements_txt = req_dep_line
            click.echo("* %s created. Install it with:\npip3 install -r %s" % (fmt.bold(REQUIREMENTS_TXT), REQUIREMENTS_TXT))

    # copy files at the very end
    for file_name in TEMPLATE_FILES + toml_files:
        shutil.copy(clone_storage.make_full_path(file_name), dest_storage.make_full_path(file_name))

    # create script
    dest_storage.save(dest_pipeline_script, dest_script_source)
    # generate tomls with comments
    secrets_prov = SecretsTomlProvider()
    write_values(secrets_prov._toml, required_secrets.values())
    config_prov = ConfigTomlProvider()
    write_values(config_prov._toml, required_config.values())
    # write toml files
    secrets_prov._write_toml()
    config_prov._write_toml()

    if requirements_txt is not None:
        dest_storage.save(REQUIREMENTS_TXT, requirements_txt)


class CliCommandException(DltException):
    def __init__(self, cmd: str, msg: str, inner_exc: Exception = None) -> None:
        self.cmd = cmd
        self.inner_exc = inner_exc
        super().__init__(msg)



import os
import tempfile
from typing import Any, Iterator, TypedDict, cast, Union

import dlt
from dlt.common import json
from dlt.common.configuration.specs import BaseConfiguration
from dlt.common.runners.venv import Venv
from dlt.common.typing import DictStrAny, StrAny, StrOrBytesPath, TDataItem, TDataItems

from examples.sources.stdout import json_stdout as singer_process_pipe

FilePathOrDict = Union[StrAny, StrOrBytesPath]

class SingerMessage(TypedDict):
    type: str  # noqa: A003


class SingerRecord(SingerMessage):
    record: DictStrAny
    stream: str


class SingerState(SingerMessage):
    value: DictStrAny

# try:
#     from singer import parse_message_from_obj, Message, RecordMessage, StateMessage
# except ImportError:
#     raise MissingDependencyException("Singer Source", ["python-dlt-singer"], "Singer runtime compatible with DLT")


# pip install ../singer/singer-python
# https://github.com/datamill-co/singer-runner/tree/master/singer_runner
# https://techgaun.github.io/active-forks/index.html#singer-io/singer-python
def get_source_from_stream(singer_messages: Iterator[SingerMessage], state: DictStrAny = None) -> Iterator[TDataItem]:
    last_state = {}
    for msg in singer_messages:
        if msg["type"] == "RECORD":
            # yield record
            msg = cast(SingerRecord, msg)
            yield dlt.with_table_name(msg["record"], msg["stream"])
        if msg["type"] == "STATE":
            msg = cast(SingerState, msg)
            last_state = msg["value"]
    if state is not None:
        state["singer"] = last_state


@dlt.transformer()
def singer_raw_stream(singer_messages: TDataItems, use_state: bool = True) -> Iterator[TDataItem]:
    if use_state:
        state = dlt.state()
    else:
        state = None
    yield from get_source_from_stream(cast(Iterator[SingerMessage], singer_messages), state)


@dlt.source(spec=BaseConfiguration)  # use BaseConfiguration spec to prevent injections
def tap(venv: Venv, tap_name: str, config_file: FilePathOrDict, catalog_file: FilePathOrDict, use_state: bool = True) -> Any:
    # TODO: generate append/replace dispositions and some table/column hints from catalog files

    def as_config_file(config: FilePathOrDict) -> StrOrBytesPath:
        if type(config) is dict:
            fd, tmp_name = tempfile.mkstemp(dir=venv.context.env_dir)
            with os.fdopen(fd, "w") as f:
                json.dump(config, f)
            return cast(str, tmp_name)
        else:
            return config  # type: ignore

    # write config dictionary to temp file in virtual environment if passed as dict
    config_file_path = as_config_file(config_file)

    # process catalog like config
    catalog_file_path = as_config_file(catalog_file)

    @dlt.resource(name=tap_name)
    def singer_messages() -> Iterator[TDataItem]:
        # possibly pass state
        if use_state:
            state = dlt.state()
        else:
            state = None
        if state is not None and state.get("singer"):
            state_params = ("--state", as_config_file(dlt.state()["singer"]))
        else:
            state_params = ()  # type: ignore

        pipe_iterator = singer_process_pipe(venv,
                                        tap_name,
                                        "--config",
                                        os.path.abspath(config_file_path),
                                        "--catalog",
                                        os.path.abspath(catalog_file_path),
                                        *state_params
                                        )
        yield from get_source_from_stream(pipe_iterator, state)  # type: ignore

    return singer_messages



from typing import Any, Iterator, Sequence, Union, cast

import dlt
from dlt.common.configuration.specs import GcpClientCredentialsWithDefault
from dlt.common.typing import DictStrAny, StrAny
from dlt.common.exceptions import MissingDependencyException

try:
    from apiclient.discovery import build
except ImportError:
    raise MissingDependencyException("Google Sheets Source", ["google1", "google2"])


# gets sheet values from a single spreadsheet preserving original typing for schema generation
# TODO: write schema discovery to better handle dates
# TODO: consider using https://github.com/burnash/gspread for spreadsheet discovery


def _initialize_sheets(credentials: GcpClientCredentialsWithDefault) -> Any:
    # Build the service object.
    service = build('sheets', 'v4', credentials=credentials.to_service_account_credentials())

    return service


@dlt.source
def google_spreadsheet(spreadsheet_id: str, sheet_names: Sequence[str], credentials: Union[GcpClientCredentialsWithDefault, str, StrAny] = dlt.secrets.value) -> Any:

    sheets = _initialize_sheets(cast(GcpClientCredentialsWithDefault, credentials))

    def get_sheet(sheet_name: str) -> Iterator[DictStrAny]:

        # get list of list of typed values
        result = sheets.spreadsheets().values().get(
            spreadsheetId=spreadsheet_id,
            range=sheet_name,
            # unformatted returns typed values
            valueRenderOption="UNFORMATTED_VALUE",
            # will return formatted dates
            dateTimeRenderOption="FORMATTED_STRING"
        ).execute()

        values = result.get('values')

        # yield dicts assuming row 0 contains headers and following rows values and all rows have identical length
        for v in values[1:]:
            yield {h: v for h, v in zip(values[0], v)}

    # create resources from supplied sheet names
    return [dlt.resource(get_sheet(name), name=name, write_disposition="replace") for name in sheet_names]



import os
import tempfile
import datetime  # noqa: 251
from typing import Any, Callable, ClassVar, Dict, List, NamedTuple, Optional, Protocol, Sequence, Tuple, TypedDict

from dlt.common.configuration.container import ContainerInjectableContext
from dlt.common.configuration import configspec
from dlt.common.destination import TDestinationReferenceArg
from dlt.common.schema import Schema
from dlt.common.schema.typing import TColumnSchema, TWriteDisposition
from dlt.common.typing import DictStrAny


class LoadInfo(NamedTuple):
    """A tuple holding the information on recently loaded packages. Returned by pipeline run method"""
    pipeline: "SupportsPipeline"
    destination_name: str
    destination_displayable_credentials: str
    dataset_name: str
    loads_ids: Dict[str, bool]
    failed_jobs: Dict[str, Sequence[Tuple[str, str]]]

    def __str__(self) -> str:
        msg = f"{len(self.loads_ids)} load package(s) were loaded to destination {self.destination_name} and into dataset {self.dataset_name}\n"
        msg += f"The {self.destination_name} destination used {self.destination_displayable_credentials} location to store data\n"
        for load_id, completed in self.loads_ids.items():
            cstr = "COMPLETED" if completed else "NOT COMPLETED"

            # now enumerate all complete loads if we have any failed packages
            # complete but failed job will not raise any exceptions
            failed_jobs = self.failed_jobs.get(load_id)
            jobs_str = "no failed jobs\n" if not failed_jobs else f"{len(failed_jobs)} FAILED job(s)!\n"
            msg += f"Load package {load_id} is {cstr} and contains {jobs_str}"
            for job_id, failed_message in failed_jobs:
                msg += f"\t{job_id}: {failed_message}\n"
        return msg


class TPipelineState(TypedDict, total=False):
    """Schema for a pipeline state that is stored within the pipeline working directory"""
    pipeline_name: str
    dataset_name: str
    default_schema_name: Optional[str]
    """Name of the first schema added to the pipeline to which all the resources without schemas will be added"""
    schema_names: Optional[List[str]]
    """All the schemas present within the pipeline working directory"""
    destination: Optional[str]

    # properties starting with _ are not automatically applied to pipeline object when state is restored
    _state_version: int
    _state_engine_version: int
    _last_extracted_at: datetime.datetime
    """Timestamp indicating when the state was synced with the destination. Lack of timestamp means not synced state."""


class SupportsPipeline(Protocol):
    """A protocol with core pipeline operations that lets high level abstractions ie. sources to access pipeline methods and properties"""
    pipeline_name: str
    @property
    def state(self) -> TPipelineState:
        ...

    def run(
        self,
        data: Any = None,
        *,
        destination: TDestinationReferenceArg = None,
        dataset_name: str = None,
        credentials: Any = None,
        table_name: str = None,
        write_disposition: TWriteDisposition = None,
        columns: Sequence[TColumnSchema] = None,
        schema: Schema = None
    ) -> LoadInfo:
        ...


class SupportsPipelineRun(Protocol):
    def __call__(
        self,
        *,
        destination: TDestinationReferenceArg = None,
        dataset_name: str = None,
        credentials: Any = None,
        table_name: str = None,
        write_disposition: TWriteDisposition = None,
        columns: Sequence[TColumnSchema] = None,
        schema: Schema = None
    ) -> LoadInfo:
        ...


@configspec(init=True)
class PipelineContext(ContainerInjectableContext):
    _deferred_pipeline: Callable[[], SupportsPipeline]
    _pipeline: SupportsPipeline

    can_create_default: ClassVar[bool] = False

    def pipeline(self) -> SupportsPipeline:
        """Creates or returns exiting pipeline"""
        if not self._pipeline:
            # delayed pipeline creation
            self._pipeline = self._deferred_pipeline()
        return self._pipeline

    def activate(self, pipeline: SupportsPipeline) -> None:
        self._pipeline = pipeline

    def is_active(self) -> bool:
        return self._pipeline is not None

    def deactivate(self) -> None:
        self._pipeline = None

    def __init__(self, deferred_pipeline: Callable[..., SupportsPipeline]) -> None:
        """Initialize the context with a function returning the Pipeline object to allow creation on first use"""
        self._deferred_pipeline = deferred_pipeline


def get_default_working_dir() -> str:
    """ Gets default working dir of the pipeline, which may be
        1. in user home directory ~/.dlt/pipelines/
        2. if current user is root in /var/dlt/pipelines
        3. if current user does not have a home directory in /tmp/dlt/pipelines
    """
    # getuid not available on Windows
    if hasattr(os, "getuid") and os.geteuid() == 0:
        # we are root so use standard /var
        return os.path.join("/var", "dlt", "pipelines")

    home = _get_home_dir()
    if home is None:
        # no home dir - use temp
        return os.path.join(tempfile.gettempdir(), "dlt", "pipelines")
    else:
        # if home directory is available use ~/.dlt/pipelines
        return os.path.join(home, ".dlt", "pipelines")


def _get_home_dir() -> str:
    return os.path.expanduser("~")



from typing import Iterator, List, Any, Union
from functools import partial

import dlt

from dlt.common.configuration.specs.postgres_credentials import ConnectionStringCredentials
from dlt.common.typing import AnyFun, DictStrAny, StrAny, TDataItem
from dlt.common.exceptions import MissingDependencyException


try:
    # import gracefully and produce nice exception that explains the user what to do
    import pandas
except ImportError:
    raise MissingDependencyException("SQL Query Source", ["pandas"], "SQL Query Source temporarily uses pandas as DB interface")

try:
    from sqlalchemy.exc import NoSuchModuleError
except ImportError:
    raise MissingDependencyException("SQL Query Source", ["sqlalchemy"], "SQL Query Source temporarily uses pandas as DB interface")


def _query_data(
    f: AnyFun
) -> Iterator[DictStrAny]:

    try:
        items = f()
    except NoSuchModuleError as m_exc:
        if "redshift.redshift_connector" in str(m_exc):
            raise MissingDependencyException("SQL Query Source", ["sqlalchemy-redshift", "redshift_connector"], "Redshift dialect support for SqlAlchemy")
        raise

    for i in items:
        record = i.to_dict("records")
        if len(record) > 0:
            yield record[0]


@dlt.resource
def query_table(
    table_name: str,
    credentials: Union[ConnectionStringCredentials, str, StrAny],
    table_schema_name: str = None,
    # index_col: Union[str, Sequence[str], None] = None,
    coerce_float: bool = True,
    parse_dates: Any = None,
    columns: List[str] = None,
    chunk_size: int = 1000
) -> Any:
    assert isinstance(credentials, ConnectionStringCredentials)
    f = partial(pandas.read_sql_table, table_name, credentials.to_native_representation(), table_schema_name, None, coerce_float, parse_dates, columns, chunksize=chunk_size)
    # if resource is returned from decorator function, it will override the hints from decorator
    return dlt.resource(_query_data(f), name=table_name)


@dlt.resource
def query_sql(
    sql: str,
    credentials: Union[ConnectionStringCredentials, str, StrAny],
    coerce_float: bool = True,
    parse_dates: Any = None,
    chunk_size: int = 1000,
    dtype: Any = None
) -> Iterator[TDataItem]:
    assert isinstance(credentials, ConnectionStringCredentials)
    f = partial(pandas.read_sql_query, sql, credentials.to_native_representation(), None, coerce_float, None, parse_dates, chunk_size, dtype)
    yield from _query_data(f)



from typing import Optional

from dlt.common.configuration import configspec
from dlt.common.configuration.specs import RunConfiguration, BaseConfiguration
from dlt.common.typing import TSecretValue
from dlt.common.utils import digest256



@configspec
class PipelineConfiguration(BaseConfiguration):
    pipeline_name: Optional[str] = None
    working_dir: Optional[str] = None
    destination_name: Optional[str] = None
    pipeline_salt: Optional[TSecretValue] = None
    restore_from_destination: bool = False
    runtime: RunConfiguration

    def on_resolved(self) -> None:
        if not self.pipeline_name:
            self.pipeline_name = self.runtime.pipeline_name
        if not self.pipeline_salt:
            self.pipeline_salt = TSecretValue(digest256(self.pipeline_name))



import dlt
from dlt.destinations import postgres
from dlt.common.schema import utils

from examples.sources.sql_query import query_table, query_sql

# the connection string to redshift instance holding some ethereum data
# the connection string does not contain the password element and you should provide it in environment variable: SOURCES__CREDENTIALS__PASSWORD
source_dsn = "redshift+redshift_connector://loader@chat-analytics.czwteevq7bpe.eu-central-1.redshift.amazonaws.com:5439/chat_analytics_rasa"

# get data from table, we preserve method signature from pandas
items = query_table("blocks__transactions", source_dsn, table_schema_name="mainnet_2_ethereum", coerce_float=False)
# the data is also an iterator

for i in items:
    assert isinstance(i, dict)
    for k, v in i.items():
        print(f"{k}:{v} ({type(v)}:{utils.py_type_to_sc_type(type(v))})")

# get data from query
items = query_sql("select *  from mainnet_2_ethereum.blocks__transactions limit 10", source_dsn)

# and load it into a local postgres instance
# the connection string does not have the password part. provide it in DESTINATION__CREDENTIALS__PASSWORD
# you can find a docker compose file that spins up required instance in tests/load/postgres
# note: run the script without required env variables to see info on possible secret configurations that were tried

info = dlt.pipeline().run(items, destination=postgres, dataset_name="ethereum", table_name="transactions")
print(info)



from typing import Optional, Sequence, Tuple
from git import GitError
from prometheus_client import REGISTRY, Gauge, CollectorRegistry, Info
from prometheus_client.metrics import MetricWrapperBase

from dlt.common import logger
from dlt.cli import TRunnerArgs
from dlt.common.typing import DictStrAny, DictStrStr, StrAny
from dlt.common.logger import is_json_logging
from dlt.common.telemetry import get_logging_extras
from dlt.common.configuration.specs import GcpClientCredentials
from dlt.common.storages import FileStorage
from dlt.common.runners import initialize_runner, run_pool
from dlt.common.telemetry import TRunMetrics
from dlt.common.git import git_custom_key_command, ensure_remote_head, clone_repo

from dlt.dbt_runner.configuration import DBTRunnerConfiguration, gen_configuration_variant
from dlt.dbt_runner.utils import DBTProcessingError, dbt_results, initialize_dbt_logging, is_incremental_schema_out_of_sync_error, run_dbt_command
from dlt.dbt_runner.exceptions import PrerequisitesException


CLONED_PACKAGE_NAME = "dbt_package"

CONFIG: DBTRunnerConfiguration = None
storage: FileStorage = None
dbt_package_vars: StrAny = None
global_args: Sequence[str] = None
repo_path: str = None
profile_name: str = None

model_elapsed_gauge: Gauge = None
model_exec_info: Info = None


def create_folders() -> Tuple[FileStorage, StrAny, Sequence[str], str, str]:
    storage_ = FileStorage(CONFIG.package_volume_path, makedirs=True)
    dbt_package_vars_: DictStrAny = {
        "source_schema_prefix": CONFIG.source_schema_prefix
    }
    if CONFIG.dest_schema_prefix:
        dbt_package_vars_["dest_schema_prefix"] = CONFIG.dest_schema_prefix
    if CONFIG.package_additional_vars:
        dbt_package_vars_.update(CONFIG.package_additional_vars)

    # initialize dbt logging, returns global parameters to dbt command
    global_args_ = initialize_dbt_logging(CONFIG.log_level, is_json_logging(CONFIG.log_format))

    # generate path for the dbt package repo
    repo_path_ = storage_.make_full_path(CLONED_PACKAGE_NAME)

    # generate profile name
    profile_name_: str = None
    if CONFIG.package_profile_prefix:
        if isinstance(CONFIG, GcpClientCredentials):
            profile_name_ = "%s_bigquery" % (CONFIG.package_profile_prefix)
        else:
            profile_name_ = "%s_redshift" % (CONFIG.package_profile_prefix)

    return storage_, dbt_package_vars_, global_args_, repo_path_, profile_name_


def create_gauges(registry: CollectorRegistry) -> Tuple[MetricWrapperBase, MetricWrapperBase]:
    return (
        Gauge("dbtrunner_model_elapsed_seconds", "Last model processing time", ["model"], registry=registry),
        Info("dbtrunner_model_status", "Last execution status of the model", registry=registry)
    )


def run_dbt(command: str, command_args: Sequence[str] = None) -> Sequence[dbt_results. BaseResult]:
    logger.info(f"Exec dbt command: {global_args} {command} {command_args} {dbt_package_vars} on profile {profile_name or '<project_default>'}")
    return run_dbt_command(
        repo_path, command,
        CONFIG.package_profiles_dir,
        profile_name=profile_name,
        command_args=command_args,
        global_args=global_args,
        dbt_vars=dbt_package_vars
    )


def log_dbt_run_results(results: dbt_results.RunExecutionResult) -> None:
    # run may return RunResult of something different depending on error
    if issubclass(type(results), dbt_results.BaseResult):
        results = [results]  # make it iterable
    elif issubclass(type(results), dbt_results.ExecutionResult):
        pass
    else:
        logger.warning(f"{type(results)} is unknown and cannot be logged")
        return

    info: DictStrStr = {}
    for res in results:
        name = res.node.name
        message = res.message
        time = res.execution_time
        if res.status == dbt_results.RunStatus.Error:
            logger.error(f"Model {name} error! Error: {message}")
        else:
            logger.info(f"Model {name} {res.status} in {time} seconds with {message}")
        model_elapsed_gauge.labels(name).set(time)
        info[name] = message

    # log execution
    model_exec_info.info(info)
    logger.metrics("Executed models", extra=get_logging_extras([model_elapsed_gauge, model_exec_info]))


def initialize_package(with_git_command: Optional[str]) -> None:
    try:
        # cleanup package folder
        if storage.has_folder(CLONED_PACKAGE_NAME):
            storage.delete_folder(CLONED_PACKAGE_NAME, recursively=True)
        logger.info(f"Will clone {CONFIG.package_repository_url} head {CONFIG.package_repository_branch} into {repo_path}")
        clone_repo(CONFIG.package_repository_url, repo_path, branch=CONFIG.package_repository_branch, with_git_command=with_git_command)
        run_dbt("deps")
    except Exception:
        # delete folder so we start clean next time
        if storage.has_folder(CLONED_PACKAGE_NAME):
            storage.delete_folder(CLONED_PACKAGE_NAME, recursively=True)
        raise


def ensure_newest_package() -> None:
    with git_custom_key_command(CONFIG.package_repository_ssh_key) as ssh_command:
        try:
            ensure_remote_head(repo_path, with_git_command=ssh_command)
        except GitError as err:
            # cleanup package folder
            logger.info(f"Package will be cloned due to {type(err).__name__}:{str(err)}")
            initialize_package(with_git_command=ssh_command)


def run_db_steps() -> Sequence[dbt_results.BaseResult]:
    # make sure we use package from the remote head
    ensure_newest_package()
    # check if raw schema exists
    try:
        if CONFIG.package_source_tests_selector:
            run_dbt("test", ["-s", CONFIG.package_source_tests_selector])
    except DBTProcessingError as err:
        raise PrerequisitesException() from err

    # always run seeds
    run_dbt("seed")
    # throws DBTProcessingError
    try:
        return run_dbt("run", CONFIG.package_run_params)
    except DBTProcessingError as e:
        # detect incremental model out of sync
        if is_incremental_schema_out_of_sync_error(e.results) and CONFIG.auto_full_refresh_when_out_of_sync:
            logger.warning(f"Attempting full refresh due to incremental model out of sync on {e.results.message}")
            return run_dbt("run", CONFIG.package_run_params + ["--full-refresh"])
        else:
            raise


def run(_: None) -> TRunMetrics:
    try:
        # there were many issues with running the method below with pool.apply
        # 1 - some exceptions are not serialized well on process boundary and queue hangs
        # 2 - random hangs event if there's no exception, probably issues with DBT spawning its own workers
        # instead the runner host was configured to recycle each run
        results = run_db_steps()
        log_dbt_run_results(results)
        return TRunMetrics(False, False, 0)
    except PrerequisitesException:
        logger.warning("Raw schema test failed, it may yet not be created")
        # run failed and loads possibly still pending
        return TRunMetrics(False, True, 1)
    except DBTProcessingError as runerr:
        log_dbt_run_results(runerr.results)
        # pass exception to the runner
        raise


def configure(C: DBTRunnerConfiguration, collector: CollectorRegistry) -> None:
    global CONFIG
    global storage, dbt_package_vars, global_args, repo_path, profile_name
    global model_elapsed_gauge, model_exec_info

    CONFIG = C
    storage, dbt_package_vars, global_args, repo_path, profile_name = create_folders()
    try:
        model_elapsed_gauge, model_exec_info = create_gauges(REGISTRY)
    except ValueError as v:
        # ignore re-creation of gauges
        if "Duplicated" not in str(v):
            raise


def main(args: TRunnerArgs) -> int:
    C = gen_configuration_variant(args._asdict())
    # we should force single run
    initialize_runner(C)
    try:
        configure(C, REGISTRY)
    except Exception:
        logger.exception("init module")
        return -1

    return run_pool(C, run)


def run_main(args: TRunnerArgs) -> None:
    exit(main(args))



from typing import Sequence
from dlt.common.exceptions import TerminalException, TransientException, DestinationException
from dlt.common.destination import TLoadJobStatus


class DestinationTerminalException(DestinationException, TerminalException):
    pass

class DestinationTransientException(DestinationException, TransientException):
    pass


class DatabaseException(DestinationException):
    def __init__(self, dbapi_exception: Exception) -> None:
        self.dbapi_exception = dbapi_exception
        super().__init__(dbapi_exception)


class DatabaseTerminalException(DestinationTerminalException, DatabaseException):
    def __init__(self, dbapi_exception: Exception) -> None:
        super().__init__(dbapi_exception)


class DatabaseUndefinedRelation(DatabaseTerminalException):
    def __init__(self, dbapi_exception: Exception) -> None:
        super().__init__(dbapi_exception)


class DatabaseTransientException(DestinationTransientException, DatabaseException):
    def __init__(self, dbapi_exception: Exception) -> None:
        super().__init__(dbapi_exception)


class LoadClientNoConnection(DestinationTransientException):
    def __init__(self, client_type: str) -> None:
        self.client_type = client_type
        super().__init__(f"Connection in sql client {client_type} is closed. Open the connection with 'client.open_connection' or with the 'with client:' statement")


class DestinationSchemaWillNotUpdate(DestinationTerminalException):
    def __init__(self, table_name: str, columns: Sequence[str], msg: str) -> None:
        self.table_name = table_name
        self.columns = columns
        super().__init__(f"Schema for table {table_name} column(s) {columns} will not update: {msg}")


class LoadJobNotExistsException(DestinationTerminalException):
    def __init__(self, job_id: str) -> None:
        super().__init__(f"Job with id/file name {job_id} not found")


class LoadJobTerminalException(DestinationTerminalException):
    def __init__(self, file_path: str) -> None:
        super().__init__(f"Job with id/file name {file_path} encountered unrecoverable problem")


class LoadJobUnknownTableException(DestinationTerminalException):
    def __init__(self, table_name: str, file_name: str) -> None:
        self.table_name = table_name
        super().__init__(f"Client does not know table {table_name} for load file {file_name}")


class LoadJobInvalidStateTransitionException(DestinationTerminalException):
    def __init__(self, from_state: TLoadJobStatus, to_state: TLoadJobStatus) -> None:
        self.from_state = from_state
        self.to_state = to_state
        super().__init__(f"Load job cannot transition form {from_state} to {to_state}")


class LoadJobFileTooBig(DestinationTerminalException):
    def __init__(self, file_name: str, max_size: int) -> None:
        super().__init__(f"File {file_name} exceeds {max_size} and cannot be loaded. Split the file and try again.")



# from dlt.pipeline import Pipeline,  GCPPipelineCredentials

# # credentials = PostgresPipelineCredentials("redshift", "chat_analytics_rasa", "mainnet_6", "loader", "3.73.90.3")
# credentials = GCPPipelineCredentials.from_services_file("_secrets/project1234_service.json", "mainnet_4")
# pipeline = Pipeline("ethereum")
# pipeline.create_pipeline(credentials)
# schema = Pipeline.load_schema_from_file("examples/schemas/ethereum_schema.yml")
# # set the loaded schema for the whole pipeline
# pipeline.set_default_schema(schema)
# # will sync schema with the target
# # pipeline.sync_schema()
# # with pipeline.sql_client() as c:
# #     c.execute_sql("BEGIN TRANSACTION;\n INSERT INTO `mainnet_6_ethereum._dlt_loads` VALUES('2022-06-09 22:09:21.254700 UTC', '1654812536.548904', 1); COMMIT TRANSACTION;")



import os
import inspect
from types import ModuleType
from makefun import wraps
from typing import Any, Callable, Dict, Iterator, List, NamedTuple, Optional, Tuple, Type, TypeVar, Union, cast, overload

from dlt.common.configuration import with_config, get_fun_spec
from dlt.common.configuration.resolve import inject_namespace
from dlt.common.configuration.specs import BaseConfiguration
from dlt.common.configuration.specs.config_namespace_context import ConfigNamespacesContext
from dlt.common.schema.schema import Schema
from dlt.common.schema.typing import TTableSchemaColumns, TWriteDisposition
from dlt.common.storages.exceptions import SchemaNotFoundError
from dlt.common.storages.schema_storage import SchemaStorage
from dlt.common.typing import AnyFun, ParamSpec, Concatenate, TDataItem, TDataItems
from dlt.common.utils import get_callable_name, is_inner_callable
from dlt.extract.exceptions import InvalidTransformerDataTypeGeneratorFunctionRequired, ResourceFunctionExpected, SourceDataIsNone, SourceIsAClassTypeError, SourceNotAFunction

from dlt.extract.typing import TTableHintTemplate
from dlt.extract.source import DltResource, DltSource, TUnboundDltResource


class SourceInfo(NamedTuple):
    SPEC: Type[BaseConfiguration]
    f: AnyFun
    module: ModuleType


_SOURCES: Dict[str, SourceInfo] = {}

TSourceFunParams = ParamSpec("TSourceFunParams")
TResourceFunParams = ParamSpec("TResourceFunParams")


@overload
def source(func: Callable[TSourceFunParams, Any], /, name: str = None, schema: Schema = None, spec: Type[BaseConfiguration] = None) -> Callable[TSourceFunParams, DltSource]:
    ...

@overload
def source(func: None = ..., /, name: str = None, schema: Schema = None, spec: Type[BaseConfiguration] = None) -> Callable[[Callable[TSourceFunParams, Any]], Callable[TSourceFunParams, DltSource]]:
    ...

def source(func: Optional[AnyFun] = None, /, name: str = None, schema: Schema = None, spec: Type[BaseConfiguration] = None) -> Any:

    def decorator(f: Callable[TSourceFunParams, Any]) -> Callable[TSourceFunParams, DltSource]:
        nonlocal schema, name

        if not callable(f) or isinstance(f, DltResource):
            raise SourceNotAFunction(name or "<no name>", f, type(f))

        if inspect.isclass(f):
            raise SourceIsAClassTypeError(name or "<no name>", f)

        # source name is passed directly or taken from decorated function name
        name = name or get_callable_name(f)

        if not schema:
            # load the schema from file with name_schema.yaml/json from the same directory, the callable resides OR create new default schema
            schema = _maybe_load_schema_for_callable(f, name) or Schema(name, normalize_name=True)

        # wrap source extraction function in configuration with namespace
        source_namespaces = ("sources", name)
        conf_f = with_config(f, spec=spec, namespaces=source_namespaces)

        @wraps(conf_f, func_name=name)
        def _wrap(*args: Any, **kwargs: Any) -> DltSource:
            # configurations will be accessed in this namespace in the source
            with inject_namespace(ConfigNamespacesContext(namespaces=source_namespaces)):
                rv = conf_f(*args, **kwargs)

            # if generator, consume it immediately
            if inspect.isgenerator(rv):
                rv = list(rv)

            if rv is None:
                raise SourceDataIsNone(name)

            # convert to source
            return DltSource.from_data(name, schema, rv)

        # get spec for wrapped function
        SPEC = get_fun_spec(conf_f)
        # store the source information
        _SOURCES[_wrap.__qualname__] = SourceInfo(SPEC, _wrap, inspect.getmodule(f))

        # the typing is right, but makefun.wraps does not preserve signatures
        return _wrap  # type: ignore

    if func is None:
        # we're called with parens.
        return decorator

    # we're called as @source without parens.
    return decorator(func)


# @source
# def reveal_1() -> None:
#     pass

# @source(name="revel")
# def reveal_2() -> None:
#     pass


# def revel_3(v) -> int:
#     pass


# reveal_type(reveal_1)
# reveal_type(reveal_1())

# reveal_type(reveal_2)
# reveal_type(reveal_2())

# reveal_type(source(revel_3))
# reveal_type(source(revel_3)("s"))

@overload
def resource(
    data: Callable[TResourceFunParams, Any],
    /,
    name: str = None,
    table_name: TTableHintTemplate[str] = None,
    write_disposition: TTableHintTemplate[TWriteDisposition] = None,
    columns: TTableHintTemplate[TTableSchemaColumns] = None,
    selected: bool = True,
    spec: Type[BaseConfiguration] = None
) -> Callable[TResourceFunParams, DltResource]:
    ...

@overload
def resource(
    data: None = ...,
    /,
    name: str = None,
    table_name: TTableHintTemplate[str] = None,
    write_disposition: TTableHintTemplate[TWriteDisposition] = None,
    columns: TTableHintTemplate[TTableSchemaColumns] = None,
    selected: bool = True,
    spec: Type[BaseConfiguration] = None
) -> Callable[[Callable[TResourceFunParams, Any]], Callable[TResourceFunParams, DltResource]]:
    ...


# @overload
# def resource(
#     data: Union[DltSource, DltResource, Sequence[DltSource], Sequence[DltResource]],
#     /
# ) -> DltResource:
#     ...


@overload
def resource(
    data: Union[List[Any], Tuple[Any], Iterator[Any]],
    /,
    name: str = None,
    table_name: TTableHintTemplate[str] = None,
    write_disposition: TTableHintTemplate[TWriteDisposition] = None,
    columns: TTableHintTemplate[TTableSchemaColumns] = None,
    selected: bool = True,
    spec: Type[BaseConfiguration] = None
) -> DltResource:
    ...


def resource(
    data: Optional[Any] = None,
    /,
    name: str = None,
    table_name: TTableHintTemplate[str] = None,
    write_disposition: TTableHintTemplate[TWriteDisposition] = None,
    columns: TTableHintTemplate[TTableSchemaColumns] = None,
    selected: bool = True,
    spec: Type[BaseConfiguration] = None,
    depends_on: TUnboundDltResource = None
) -> Any:

    def make_resource(_name: str, _data: Any) -> DltResource:
        table_template = DltResource.new_table_template(table_name or _name, write_disposition=write_disposition, columns=columns)
        return DltResource.from_data(_data, _name, table_template, selected, cast(DltResource, depends_on))


    def decorator(f: Callable[TResourceFunParams, Any]) -> Callable[TResourceFunParams, DltResource]:
        if not callable(f):
            if depends_on:
                # raise more descriptive exception if we construct transformer
                raise InvalidTransformerDataTypeGeneratorFunctionRequired(name or "<no name>", f, type(f))
            raise ResourceFunctionExpected(name or "<no name>", f, type(f))

        resource_name = name or get_callable_name(f)

        # do not inject config values for inner functions, we assume that they are part of the source
        SPEC: Type[BaseConfiguration] = None
        if is_inner_callable(f):
            conf_f = f
        else:
            # wrap source extraction function in configuration with namespace
            conf_f = with_config(f, spec=spec, namespaces=("sources", resource_name))
            # get spec for wrapped function
            SPEC = get_fun_spec(conf_f)

        # store the standalone resource information
        if SPEC:
            _SOURCES[f.__qualname__] = SourceInfo(SPEC, f, inspect.getmodule(f))

        return make_resource(resource_name, conf_f)

    # if data is callable or none use decorator
    if data is None:
        # we're called with parens.
        return decorator

    if callable(data):
        return decorator(data)
    else:
        # take name from the generator
        if inspect.isgenerator(data):
            name = name or get_callable_name(data)  # type: ignore
        return make_resource(name, data)


def transformer(
    data_from: TUnboundDltResource = DltResource.Empty,
    name: str = None,
    table_name: TTableHintTemplate[str] = None,
    write_disposition: TTableHintTemplate[TWriteDisposition] = None,
    columns: TTableHintTemplate[TTableSchemaColumns] = None,
    selected: bool = True,
    spec: Type[BaseConfiguration] = None
) -> Callable[[Callable[Concatenate[TDataItem, TResourceFunParams], Any]], Callable[TResourceFunParams, DltResource]]:
    f: AnyFun = None
    # if data_from is a function we are called without parens
    if inspect.isfunction(data_from):
        f = data_from
        data_from = DltResource.Empty
    return resource(f, name=name, table_name=table_name, write_disposition=write_disposition, columns=columns, selected=selected, spec=spec, depends_on=data_from)  # type: ignore


def _maybe_load_schema_for_callable(f: AnyFun, name: str) -> Optional[Schema]:
    if not inspect.isfunction(f):
        f = f.__class__
    try:
        file = inspect.getsourcefile(f)
        if file:
            return SchemaStorage.load_schema_file(os.path.dirname(file), name)
    except SchemaNotFoundError:
        pass
    return None


def _get_source_for_inner_function(f: AnyFun) -> Optional[SourceInfo]:
    # find source function
    parts = get_callable_name(f, "__qualname__").split(".")
    parent_fun = ".".join(parts[:-2])
    return _SOURCES.get(parent_fun)

# @resource
# def reveal_1() -> None:
#     yield 1


# @transformer(data_from=reveal_1)
# def transf(item, a, b, c=None):
#     yield item

# reveal_type(transf)


# @resource(name="revel")
# def reveal_2() -> None:
#     pass


# def revel_3(v) -> int:
#     pass


# reveal_type(reveal_1)
# reveal_type(reveal_1())

# reveal_type(reveal_2)
# reveal_type(reveal_2())

# reveal_type(resource(revel_3))
# reveal_type(resource(revel_3)("s"))


# reveal_type(resource([], name="aaaa"))
# reveal_type(resource("aaaaa", name="aaaa"))

# name of dlt metadata as part of the item
# DLT_METADATA_FIELD = "_dlt_meta"


# class TEventDLTMeta(TypedDict, total=False):
#     table_name: str  # a root table in which store the event


# def append_dlt_meta(item: TBoundItem, name: str, value: Any) -> TBoundItem:
#     if isinstance(item, abc.Sequence):
#         for i in item:
#             i.setdefault(DLT_METADATA_FIELD, {})[name] = value
#     elif isinstance(item, dict):
#         item.setdefault(DLT_METADATA_FIELD, {})[name] = value

#     return item


# def with_table_name(item: TBoundItem, table_name: str) -> TBoundItem:
#     # normalize table name before adding
#     return append_dlt_meta(item, "table_name", table_name)


# def get_table_name(item: StrAny) -> Optional[str]:
#     if DLT_METADATA_FIELD in item:
#         meta: TEventDLTMeta = item[DLT_METADATA_FIELD]
#         return meta.get("table_name", None)
#     return None


# def with_retry(max_retries: int = 3, retry_sleep: float = 1.0) -> Callable[[Callable[_TFunParams, TBoundItem]], Callable[_TFunParams, TBoundItem]]:

#     def decorator(f: Callable[_TFunParams, TBoundItem]) -> Callable[_TFunParams, TBoundItem]:

#         def _wrap(*args: Any, **kwargs: Any) -> TBoundItem:
#             attempts = 0
#             while True:
#                 try:
#                     return f(*args, **kwargs)
#                 except Exception as exc:
#                     if attempts == max_retries:
#                         raise
#                     attempts += 1
#                     logger.warning(f"Exception {exc} in iterator, retrying {attempts} / {max_retries}")
#                     sleep(retry_sleep)

#         return _wrap

#     return decorator


TBoundItems = TypeVar("TBoundItems", bound=TDataItems)
TDeferred = Callable[[], TBoundItems]
TDeferredFunParams = ParamSpec("TDeferredFunParams")


def defer(f: Callable[TDeferredFunParams, TBoundItems]) -> Callable[TDeferredFunParams, TDeferred[TBoundItems]]:

    @wraps(f)
    def _wrap(*args: Any, **kwargs: Any) -> TDeferred[TBoundItems]:
        def _curry() -> TBoundItems:
            return f(*args, **kwargs)
        return _curry

    return _wrap  # type: ignore



from abc import abstractmethod
import platform

from dlt.common.wei import EVM_DECIMAL_PRECISION
if platform.python_implementation() == "PyPy":
    import psycopg2cffi as psycopg2
    from psycopg2cffi.sql import SQL, Composed
else:
    import psycopg2
    from psycopg2.sql import SQL, Composed


from typing import Any, Dict, List, Optional, Sequence, Tuple, Type

from dlt.common.arithmetics import DEFAULT_NUMERIC_PRECISION, DEFAULT_NUMERIC_SCALE
from dlt.common.destination import DestinationCapabilitiesContext, DestinationClientDwhConfiguration, LoadJob, TLoadJobStatus
from dlt.common.schema import COLUMN_HINTS, TColumnSchema, TColumnSchemaBase, TDataType, TColumnHint, Schema, TTableSchemaColumns, add_missing_hints
from dlt.common.schema.typing import TTableSchema, TWriteDisposition
from dlt.common.storages.file_storage import FileStorage

from dlt.destinations.exceptions import DestinationSchemaWillNotUpdate
from dlt.destinations.sql_client import SqlClientBase
from dlt.destinations.job_client_impl import SqlJobClientBase, LoadEmptyJob

from dlt.destinations.postgres import capabilities
from dlt.destinations.postgres.sql_client import Psycopg2SqlClient
from dlt.destinations.postgres.configuration import PostgresClientConfiguration


SCT_TO_PGT: Dict[TDataType, str] = {
    "complex": "varchar",
    "text": "varchar",
    "double": "double precision",
    "bool": "boolean",
    "timestamp": "timestamp with time zone",
    "bigint": "bigint",
    "binary": "bytea",
    "decimal": f"numeric({DEFAULT_NUMERIC_PRECISION},{DEFAULT_NUMERIC_SCALE})"
}

PGT_TO_SCT: Dict[str, TDataType] = {
    "varchar": "text",
    "double precision": "double",
    "boolean": "bool",
    "timestamp with time zone": "timestamp",
    "bigint": "bigint",
    "bytea": "binary",
    "numeric": "decimal"
}

HINT_TO_POSTGRES_ATTR: Dict[TColumnHint, str] = {
    "unique": "UNIQUE"
}


class InsertValuesLoadJob(LoadJob):
    def __init__(self, table_name: str, write_disposition: TWriteDisposition, file_path: str, sql_client: SqlClientBase["psycopg2.connection"]) -> None:
        super().__init__(FileStorage.get_file_name_from_file_path(file_path))
        self._sql_client = sql_client
        # insert file content immediately
        self._insert(sql_client.make_qualified_table_name(table_name), write_disposition, file_path)

    def status(self) -> TLoadJobStatus:
        # this job is always done
        return "completed"

    def file_name(self) -> str:
        return self._file_name

    def exception(self) -> str:
        # this part of code should be never reached
        raise NotImplementedError()

    def _insert(self, qualified_table_name: str, write_disposition: TWriteDisposition, file_path: str) -> None:
        # TODO: implement tracking of jobs in storage, both completed and failed
        # WARNING: maximum redshift statement is 16MB https://docs.aws.amazon.com/redshift/latest/dg/c_redshift-sql.html
        # the procedure below will split the inserts into 4MB packs
        with open(file_path, "r", encoding="utf-8") as f:
            header = f.readline()
            values_mark = f.readline()
            # properly formatted file has a values marker at the beginning
            assert values_mark == "VALUES\n"
            # begin the transaction
            insert_sql = [SQL("BEGIN TRANSACTION;")]
            if write_disposition == "replace":
                insert_sql.append(SQL("DELETE FROM {};").format(SQL(qualified_table_name)))
            # is_eof = False
            while content := f.read(PostgresClient.capabilities()["max_query_length"] // 2):
                # read one more line in order to
                # 1. complete the content which ends at "random" position, not an end line
                # 2. to modify it's ending without a need to re-allocating the 8MB of "content"
                until_nl = f.readline().strip("\n")
                # write INSERT
                insert_sql.extend(
                    [SQL(header).format(SQL(qualified_table_name)),
                    SQL(values_mark),
                    SQL(content)]
                )
                # are we at end of file
                is_eof = len(until_nl) == 0 or until_nl[-1] == ";"
                if not is_eof:
                    # replace the "," with ";"
                    until_nl = until_nl[:-1] + ";\n"
                # actually this may be empty if we were able to read a full file into content
                if until_nl:
                    insert_sql.append(SQL(until_nl))
                if not is_eof:
                    # execute chunk of insert
                    self._sql_client.execute_sql(Composed(insert_sql))
                    insert_sql.clear()

        # on EOF add COMMIT TX and execute
        insert_sql.append(SQL("COMMIT TRANSACTION;"))
        self._sql_client.execute_sql(Composed(insert_sql))


class PostgresClientBase(SqlJobClientBase):

    def initialize_storage(self, wipe_data: bool = False) -> None:
        if wipe_data:
            raise NotImplementedError()
        if not self.sql_client.has_dataset():
            self.sql_client.create_dataset()

    def restore_file_load(self, file_path: str) -> LoadJob:
        # always returns completed jobs as InsertValuesLoadJob is executed
        # atomically in start_file_load so any jobs that should be recreated are already completed
        # in case of bugs in loader (asking for jobs that were never created) we are not able to detect that
        return LoadEmptyJob.from_file_path(file_path, "completed")

    def start_file_load(self, table: TTableSchema, file_path: str) -> LoadJob:
        # this is using sql_client internally and will raise a right exception
        return InsertValuesLoadJob(table["name"], table["write_disposition"], file_path, self.sql_client)

    def _get_table_update_sql(self, table_name: str, new_columns: Sequence[TColumnSchema], exists: bool) -> str:
        # build sql
        canonical_name = self.sql_client.make_qualified_table_name(table_name)
        sql = "BEGIN TRANSACTION;\n"
        if not exists:
            # build CREATE
            sql += f"CREATE TABLE {canonical_name} (\n"
            sql += ",\n".join([self._get_column_def_sql(c) for c in new_columns])
            sql += ");"
        else:
            # build ALTER as separate statement for each column (redshift limitation)
            sql += "\n".join([f"ALTER TABLE {canonical_name}\nADD COLUMN {self._get_column_def_sql(c)};" for c in new_columns])
        # scan columns to get hints
        if exists:
            # no hints may be specified on added columns
            for hint in COLUMN_HINTS:
                if any(c.get(hint, False) is True for c in new_columns):
                    hint_columns = [c["name"] for c in new_columns if c.get(hint, False)]
                    raise DestinationSchemaWillNotUpdate(canonical_name, hint_columns, f"{hint} requested after table was created")
        # TODO: add FK relations
        sql += "\nCOMMIT TRANSACTION;"
        return sql

    # TODO: implement indexes and primary keys for postgres
    def _get_in_table_constraints_sql(self, t: TTableSchema) -> str:
        # get primary key
        pass

    def _get_out_table_constrains_sql(self, t: TTableSchema) -> str:
        # set non unique indexes
        pass

    @abstractmethod
    def _get_column_def_sql(self, c: TColumnSchema) -> str:
        pass

    def get_storage_table(self, table_name: str) -> Tuple[bool, TTableSchemaColumns]:
        schema_table: TTableSchemaColumns = {}
        query = f"""
                SELECT column_name, data_type, is_nullable, numeric_precision, numeric_scale
                    FROM INFORMATION_SCHEMA.COLUMNS
                WHERE table_schema = '{self.sql_client.fully_qualified_dataset_name()}' AND table_name = '{table_name}'
                ORDER BY ordinal_position;
                """
        rows = self.sql_client.execute_sql(query)
        # if no rows we assume that table does not exist
        if len(rows) == 0:
            # TODO: additionally check if table exists
            return False, schema_table
        # TODO: pull more data to infer indexes, PK and uniques attributes/constraints
        for c in rows:
            schema_c: TColumnSchemaBase = {
                "name": c[0],
                "nullable": self._null_to_bool(c[2]),
                "data_type": self._pq_t_to_sc_t(c[1], c[3], c[4]),
            }
            schema_table[c[0]] = add_missing_hints(schema_c)
        return True, schema_table

    @staticmethod
    def _null_to_bool(v: str) -> bool:
        if v == "NO":
            return False
        elif v == "YES":
            return True
        raise ValueError(v)

    @staticmethod
    def _gen_not_null(v: bool) -> str:
        return "NOT NULL" if not v else ""

    @staticmethod
    @abstractmethod
    def _sc_t_to_pq_t(sc_t: TDataType) -> str:
        pass

    @staticmethod
    @abstractmethod
    def _pq_t_to_sc_t(pq_t: str, precision: Optional[int], scale: Optional[int]) -> TDataType:
        pass


class PostgresClient(PostgresClientBase):

    def __init__(self, schema: Schema, config: PostgresClientConfiguration) -> None:
        sql_client = Psycopg2SqlClient(
            schema.normalize_make_dataset_name(config.dataset_name, config.default_schema_name, schema.name),
            config.credentials
        )
        super().__init__(schema, config, sql_client)
        self.config: PostgresClientConfiguration = config
        self.sql_client = sql_client
        self.caps = self.capabilities()

    def _get_column_def_sql(self, c: TColumnSchema) -> str:
        hints_str = " ".join(HINT_TO_POSTGRES_ATTR.get(h, "") for h in HINT_TO_POSTGRES_ATTR.keys() if c.get(h, False) is True)
        column_name = self.caps.escape_identifier(c["name"])
        return f"{column_name} {self._sc_t_to_pq_t(c['data_type'])} {hints_str} {self._gen_not_null(c['nullable'])}"

    @staticmethod
    def _sc_t_to_pq_t(sc_t: TDataType) -> str:
        if sc_t == "wei":
            return f"numeric({2*EVM_DECIMAL_PRECISION}, {EVM_DECIMAL_PRECISION})"
        return SCT_TO_PGT[sc_t]

    @staticmethod
    def _pq_t_to_sc_t(pq_t: str, precision: Optional[int], scale: Optional[int]) -> TDataType:
        if pq_t == "numeric":
            if precision == 2*EVM_DECIMAL_PRECISION and scale == EVM_DECIMAL_PRECISION:
                return "wei"
        return PGT_TO_SCT.get(pq_t, "text")

    @classmethod
    def capabilities(cls) -> DestinationCapabilitiesContext:
        return capabilities()



import os

import dlt
from dlt.destinations import bigquery, postgres

from examples.sources.jsonl import jsonl_files
from examples.sources.rasa import rasa

from examples._helpers import pub_bigquery_credentials

# let's load to bigquery, here we provide the credentials for our public project
# credentials = pub_bigquery_credentials
credentials = "postgres://loader@localhost:5432/dlt_data"


# this is also example of resource pipelining: one resource can be source of data for another one that we call "transformer"
# in case of rasa the base source can be a file, database table (see sql_query.py), kafka topic, rabbitmq queue etc. corresponding to store or broker type

# for the simplicity let's use jsonl source to read all files with events in a directory
event_files = jsonl_files([file for file in os.scandir("examples/data/rasa_trackers")])

info = dlt.pipeline(
    full_refresh=True,
    destination=postgres,
    # export_schema_path=...  # uncomment to see the final schema in the folder you want
).run(
    rasa(event_files, store_last_timestamp=True),  # also store last timestamp so we have no duplicate events
    credentials=credentials # if you skip this parameter, the credentials will be injected by the config providers
    )

print(info)

# uncomment to see the final schema
# print(dlt.pipeline().default_schema.to_pretty_yaml())



import contextlib
from typing import TYPE_CHECKING, Any, ClassVar, Dict, Optional, cast

import dlt

from dlt.common import json
from dlt.common.configuration.container import Container
from dlt.common.configuration.exceptions import ContextDefaultCannotBeCreated
from dlt.common.configuration.specs import ContainerInjectableContext
from dlt.common.configuration.specs.base_configuration import configspec
from dlt.common.configuration.specs.config_namespace_context import ConfigNamespacesContext
from dlt.common.pipeline import PipelineContext, TPipelineState
from dlt.common.typing import DictStrAny
from dlt.common.schema.typing import LOADS_TABLE_NAME, TTableSchemaColumns
from dlt.destinations.exceptions import DatabaseUndefinedRelation
from dlt.destinations.sql_client import SqlClientBase

from dlt.extract.source import DltResource

from dlt.pipeline.exceptions import PipelineStateEngineNoUpgradePathException, PipelineStateNotAvailable


# allows to upgrade state when restored with a new version of state logic/schema
STATE_ENGINE_VERSION = 1
# state table name
STATE_TABLE_NAME = "_dlt_pipeline_state"
# state table columns
STATE_TABLE_COLUMNS: TTableSchemaColumns = {
    "version": {
        "name": "version",
        "data_type": "bigint",
        "nullable": False
    },
    "engine_version": {
        "name": "engine_version",
        "data_type": "bigint",
        "nullable": False
    },
    "pipeline_name": {
        "name": "pipeline_name",
        "data_type": "text",
        "nullable": False
    },
    "state": {
        "name": "state",
        "data_type": "text",
        "nullable": False
    },
    "created_at": {
        "name": "created_at",
        "data_type": "timestamp",
        "nullable": False
    }
}


class TSourceState(TPipelineState):
    sources: Dict[str, Dict[str, Any]]


@configspec(init=True)
class StateInjectableContext(ContainerInjectableContext):
    state: TPipelineState

    can_create_default: ClassVar[bool] = False

    if TYPE_CHECKING:
        def __init__(self, state: TPipelineState = None) -> None:
            ...


def merge_state_if_changed(old_state: TPipelineState, new_state: TPipelineState) -> Optional[TPipelineState]:
    # load state from storage to be merged with pipeline changes, currently we assume no parallel changes
    if json.dumps(old_state, sort_keys=True) == json.dumps(new_state, sort_keys=True):
        return None
    # TODO: we should probably update smarter ie. recursively
    old_state.update(new_state)
    old_state["_state_version"] += 1
    return old_state


def state_resource(state: TPipelineState) -> DltResource:
    state_doc = {
        "version": state["_state_version"],
        "engine_version": state["_state_engine_version"],
        "pipeline_name": state["pipeline_name"],
        "state": json.dumps(state),
        "created_at": state["_last_extracted_at"]
    }

    return dlt.resource([state_doc], name=STATE_TABLE_NAME, write_disposition="append", columns=STATE_TABLE_COLUMNS)


def load_state_from_destination(pipeline_name: str, sql_client: SqlClientBase[Any]) -> TPipelineState:
    try:
        query = f"SELECT state FROM {STATE_TABLE_NAME} AS s JOIN {LOADS_TABLE_NAME} AS l ON l.load_id = s._dlt_load_id WHERE pipeline_name = %s AND l.status = 0 ORDER BY created_at DESC"
        with sql_client.execute_query(query, pipeline_name) as cur:
            row = cur.fetchone()
        if not row:
            return None
        state = cast(TPipelineState, json.loads(row[0]))
        # check state engine
        if state["_state_engine_version"] != STATE_ENGINE_VERSION:
            raise PipelineStateEngineNoUpgradePathException(pipeline_name, state["_state_engine_version"], state["_state_engine_version"], STATE_ENGINE_VERSION)
        return state
    except DatabaseUndefinedRelation:
        # state will not load and will not be available
        return None


def state() -> DictStrAny:
    """Returns a dictionary with the current source state. Any JSON-serializable values can be written and the read from the state.
    The state is persisted after the data is successfully read from the source.
    """
    global _last_full_state

    container = Container()
    # get the source name from the namespace context
    source_name: str = None
    with contextlib.suppress(ContextDefaultCannotBeCreated):
        namespaces = container[ConfigNamespacesContext].namespaces
        if namespaces and len(namespaces) > 1 and namespaces[0] == "sources":
            source_name = namespaces[1]
    try:
        # get managed state that is read/write
        state: TSourceState = container[StateInjectableContext].state  # type: ignore
    except ContextDefaultCannotBeCreated:
        # check if there's pipeline context
        proxy = container[PipelineContext]
        if not proxy.is_active():
            raise PipelineStateNotAvailable(source_name)
        else:
            # get unmanaged state that is read only
            state = proxy.pipeline().state  # type: ignore

    source_state = state.setdefault("sources", {})
    if source_name:
        source_state = source_state.setdefault(source_name, {})

    # allow inspection of last returned full state
    _last_full_state = state
    return source_state

_last_full_state: TPipelineState = None



import inspect
import ast
import astunparse
from ast import NodeVisitor
from typing import Any, Dict, List
from dlt.common.reflection.utils import get_outer_func_def


import dlt.reflection.names as n


class PipelineScriptVisitor(NodeVisitor):

    def __init__(self, source: str):
        self.source = source

        self.mod_aliases: Dict[str, str] = {}
        self.func_aliases: Dict[str, str] = {}
        # self.source_aliases: Dict[str, str] = {}
        self.is_destination_imported: bool = False
        self.known_calls: Dict[str, List[inspect.BoundArguments]] = {}
        self.known_sources: Dict[str, ast.FunctionDef] = {}
        self.known_source_calls: Dict[str, List[ast.Call]] = {}

    def visit_Import(self, node: ast.Import) -> Any:
        # reflect on imported modules
        for alias in node.names:
            # detect dlt import
            if alias.name == n.DLT:
                eff_name = alias.asname or alias.name
                self.mod_aliases[eff_name] = alias.name
                self._add_f_aliases(eff_name)
            if alias.name.startswith(f"{n.DLT}.") and alias.asname is None:
                # this also imports dlt
                self.mod_aliases[alias.name] = alias.name
                self._add_f_aliases(alias.name)
            if alias.name.startswith(f"{n.DESTINATIONS}."):
                self.is_destination_imported = True
        super().generic_visit(node)

    def visit_ImportFrom(self, node: ast.ImportFrom) -> Any:
        # reflect on pipeline functions and decorators
        if node.module == n.DLT:
            for alias in node.names:
                if alias.name in n.DETECTED_FUNCTIONS:
                    self.func_aliases[alias.asname or alias.name] = alias.name
        if node.module == n.DESTINATIONS:
            self.is_destination_imported = True
        super().generic_visit(node)

    def visit_FunctionDef(self, node: ast.FunctionDef) -> Any:
        # find all sources and resources by inspecting decorators
        for deco in node.decorator_list:
            # decorators can be function calls, attributes or names
            if isinstance(deco, (ast.Name, ast.Attribute)):
                alias_name = astunparse.unparse(deco).strip()
            elif isinstance(deco, ast.Call):
                alias_name = astunparse.unparse(deco.func).strip()
            else:
                raise ValueError(self.source_segment(deco), type(deco), "Unknown decorator form")
            fn = self.func_aliases.get(alias_name)
            if fn in [n.SOURCE, n.RESOURCE]:
                self.known_sources[str(node.name)] = node
        super().generic_visit(node)

    def visit_Call(self, node: ast.Call) -> Any:
        # check if this is a call to any of known functions
        alias_name = astunparse.unparse(node.func).strip()
        fn = self.func_aliases.get(alias_name)
        if not fn:
            # try a fallback to "run" function that may be called on pipeline or source
            if isinstance(node.func, ast.Attribute) and node.func.attr == n.RUN:
                fn = n.RUN
        if fn:
            # set parent to the outer function
            node.parent = get_outer_func_def(node)  # type: ignore
            sig = n.SIGNATURES[fn]
            try:
                # bind the signature where the argument values are the corresponding ast nodes
                bound_args = sig.bind(*node.args, **{str(kwd.arg):kwd.value for kwd in node.keywords})
                bound_args.apply_defaults()
                # print(f"ALIAS: {alias_name} of {self.func_aliases.get(alias_name)} with {bound_args}")
                fun_calls = self.known_calls.setdefault(fn, [])
                fun_calls.append(bound_args)
            except TypeError:
                # skip the signature
                pass
        else:
            # check if this is a call to any known source
            if alias_name in self.known_sources:
                # set parent to the outer function
                node.parent = get_outer_func_def(node)  # type: ignore
                source_calls = self.known_source_calls.setdefault(alias_name, [])
                source_calls.append(node)

        # visit the children
        super().generic_visit(node)

    def source_segment(self, node: ast.AST) -> str:
        # TODO: this must cache parsed source. right now the full source is tokenized on every call
        return ast.get_source_segment(self.source, node)

    def _add_f_aliases(self, module_name: str) -> None:
        for fn in n.DETECTED_FUNCTIONS:
            self.func_aliases[f"{module_name}.{fn}"] = fn



from typing import Any, Iterator

import dlt
import requests

from dlt.common.typing import TDataItems


@dlt.source
def chess(chess_url: str, title: str = "GM", max_players: int = 2, year: int = 2022, month: int = 10) -> Any:

    @dlt.resource(write_disposition="replace")
    def players() -> Iterator[TDataItems]:
        # https://api.chess.com/pub/titled/{title-abbrev}
        r = requests.get(f"{chess_url}titled/{title}")
        r.raise_for_status()
        # return players one by one, you could also return a list that would be faster but there's more code
        for p in r.json()["players"][:max_players]:
            yield p

    # this resource takes data from players and returns profiles
    @dlt.transformer(data_from=players, write_disposition="replace")
    def player_profile(username: Any) -> Iterator[TDataItems]:
        r = requests.get(f"{chess_url}player/{username}")
        r.raise_for_status()
        yield r.json()

    # this resource takes data from players and returns games for the last month if not specified otherwise
    @dlt.transformer(data_from=players, write_disposition="append")
    def player_games(username: Any) -> Iterator[TDataItems]:
        # https://api.chess.com/pub/player/{username}/games/{YYYY}/{MM}
        r = requests.get(f"{chess_url}player/{username}/games/{year:04d}/{month:02d}")
        r.raise_for_status()
        yield r.json()["games"]

    return players(), player_profile, player_games

print("You must run this from the examples/chess folder")
# chess_url in config.toml, credentials for postgres in secrets.toml, credentials always under credentials key
# mind the full_refresh: it makes the pipeline to load to a distinct dataset each time it is run and always is resetting the schema and state
info = dlt.pipeline(
    destination="postgres",
    dataset_name="chess",
    full_refresh=True
).run(
    chess(dlt.config.value, max_players=5, month=9)
)
# display where the data went
print(info)


import dlt

from dlt.common.storages.schema_storage import SchemaStorage

from examples.sources.singer_tap import singer_raw_stream
from examples.sources.jsonl import jsonl_file


# load hubspot schema stub - it converts all field names with `timestamp` into timestamp type
schema = SchemaStorage.load_schema_file("examples/schemas/", "hubspot", ("yaml", ))

p = dlt.pipeline(destination="postgres", full_refresh=True)
# now load a pipeline created from jsonl resource that feeds messages into singer tap transformer
pipe = jsonl_file("examples/data/singer_taps/tap_hubspot.jsonl") | singer_raw_stream()
# provide hubspot schema
info = p.run(pipe, schema=schema, credentials="postgres://loader@localhost:5432/dlt_data")
print(info)


from dlt.pipeline import Pipeline
from dlt.common.exceptions import MissingDependencyException
from dlt.helpers.pandas import query_results_to_df, pd

try:
    import streamlit as st
    # from streamlit import SECRETS_FILE_LOC, secrets
except ImportError:
    raise MissingDependencyException("DLT Streamlit Helpers", ["streamlit"], "DLT Helpers for Streamlit should be run within a streamlit app.")


# def restore_pipeline() -> Pipeline:
#     """Restores Pipeline instance and associated credentials from Streamlit secrets

#         Current implementation requires that pipeline working dir is available at the location saved in secrets.

#     Raises:
#         PipelineBackupNotFound: Raised when pipeline backup is not available
#         CannotRestorePipelineException: Raised when pipeline working dir is not found or invalid

#     Returns:
#         Pipeline: Instance of pipeline with attached credentials
#     """
#     if "dlt" not in secrets:
#         raise PipelineException("You must backup pipeline to Streamlit first")
#     dlt_cfg = secrets["dlt"]
#     credentials = deepcopy(dict(dlt_cfg["destination"]))
#     if "default_schema_name" in credentials:
#         del credentials["default_schema_name"]
#     credentials.update(dlt_cfg["credentials"])
#     pipeline = Pipeline(dlt_cfg["pipeline_name"])
#     pipeline.restore_pipeline(credentials_from_dict(credentials), dlt_cfg["working_dir"])
#     return pipeline


# def backup_pipeline(pipeline: Pipeline) -> None:
#     """Backups pipeline state to the `secrets.toml` of the Streamlit app.

#     Pipeline credentials and working directory will be added to the Streamlit `secrets` file. This allows to access query the data loaded to the destination and
#     access definitions of the inferred schemas. See `restore_pipeline` and `write_data_explorer_page` functions in the same module.

#     Args:
#         pipeline (Pipeline): Pipeline instance, typically restored with `restore_pipeline`
#     """
#     # save pipeline state to project .config
#     # config_file_name = file_util.get_project_streamlit_file_path("config.toml")

#     # save credentials to secrets
#     if os.path.isfile(SECRETS_FILE_LOC):
#         with open(SECRETS_FILE_LOC, "r", encoding="utf-8") as f:
#             # use whitespace preserving parser
#             secrets_ = tomlkit.load(f)
#     else:
#         secrets_ = tomlkit.document()

#     # save general settings
#     secrets_["dlt"] = {
#         "working_dir": pipeline.working_dir,
#         "pipeline_name": pipeline.pipeline_name
#     }

#     # get client config
#     # TODO: pipeline api v2 should provide a direct method to get configurations
#     CONFIG: BaseConfiguration = pipeline._loader_instance.load_client_cls.CONFIG  # type: ignore
#     CREDENTIALS: CredentialsConfiguration = pipeline._loader_instance.load_client_cls.CREDENTIALS  # type: ignore

#     # save client config
#     # print(dict_remove_nones_in_place(CONFIG.as_dict(lowercase=False)))
#     dlt_c = cast(TomlContainer, secrets_["dlt"])
#     dlt_c["destination"] = dict_remove_nones_in_place(dict(CONFIG))
#     dlt_c["credentials"] = dict_remove_nones_in_place(dict(CREDENTIALS))

#     with open(SECRETS_FILE_LOC, "w", encoding="utf-8") as f:
#         # use whitespace preserving parser
#         tomlkit.dump(secrets_, f)


# def write_load_status_page(pipeline: Pipeline) -> None:



def write_data_explorer_page(pipeline: Pipeline, schema_name: str = None, show_dlt_tables: bool = False, example_query: str = "", show_charts: bool = True) -> None:
    """Writes Streamlit app page with a schema and live data preview.

    Args:
        pipeline (Pipeline): Pipeline instance to use.
        schema_name (str, optional): Name of the schema to display. If None, default schema is used.
        show_dlt_tables (bool, optional): Should show DLT internal tables. Defaults to False.
        example_query (str, optional): Example query to be displayed in the SQL Query box.
        show_charts (bool, optional): Should automatically show charts for the queries from SQL Query box. Defaults to True.

    Raises:
        MissingDependencyException: Raised when a particular python dependency is not installed
    """
    @st.experimental_memo(ttl=600)
    def run_query(query: str) -> pd.DataFrame:
        # dlt pipeline exposes configured sql client that (among others) let's you make queries against the warehouse
        with pipeline.sql_client(schema_name) as client:
            df = query_results_to_df(client, query)
            return df

    if schema_name:
        schema = pipeline.schemas[schema_name]
    else:
        schema = pipeline.default_schema
    st.title(f"Available tables in {schema.name} schema")
    # st.text(schema.to_pretty_yaml())

    for table in schema.all_tables(with_dlt_tables=show_dlt_tables):
        table_name = table["name"]
        st.header(table_name)
        if "description" in table:
            st.text(table["description"])
        if "parent" in table:
            st.text("Parent table: " + table["parent"])

        # table schema contains various hints (like clustering or partition options) that we do not want to show in basic view
        essentials_f = lambda c: {k:v for k, v in c.items() if k in ["name", "data_type", "nullable"]}

        st.table(map(essentials_f, table["columns"].values()))
        # add a button that when pressed will show the full content of a table
        if st.button("SHOW DATA", key=table_name):
            st.text(f"Full {table_name} table content")
            st.dataframe(run_query(f"SELECT * FROM {table_name}"))

    st.title("Run your query")
    sql_query = st.text_area("Enter your SQL query", value=example_query)
    if st.button("Run Query"):
        if sql_query:
            st.text("Results of a query")
            try:
                # run the query from the text area
                df = run_query(sql_query)
                # and display the results
                st.dataframe(df)

                try:
                    # now if the dataset has supported shape try to display the bar or altair chart
                    if df.dtypes.shape[0] == 1 and show_charts:
                        # try barchart
                        st.bar_chart(df)
                    if df.dtypes.shape[0] == 2 and show_charts:

                        # try to import altair charts
                        try:
                            import altair as alt
                        except ImportError:
                            raise MissingDependencyException(
                                "DLT Streamlit Helpers",
                                ["altair"],
                                "DLT Helpers for Streamlit should be run within a streamlit app."
                            )

                        # try altair
                        bar_chart = alt.Chart(df).mark_bar().encode(
                            x=f'{df.columns[1]}:Q',
                            y=alt.Y(f'{df.columns[0]}:N', sort='-x')
                        )
                        st.altair_chart(bar_chart, use_container_width=True)
                except Exception as ex:
                    st.error(f"Chart failed due to: {ex}")
            except Exception as ex:
                st.text("Exception when running query")
                st.exception(ex)



import yaml
import os
import argparse
import click
from typing import Callable

from dlt.common import json
from dlt.common.schema import Schema
from dlt.common.typing import DictStrAny

from dlt.pipeline import attach

from dlt.cli import TRunnerArgs
from dlt.cli.init_command import init_command


def add_pool_cli_arguments(parser: argparse.ArgumentParser) -> None:
    parser.add_argument("--is-single-run", action="store_true", help="exit when all pending items are processed")
    parser.add_argument("--wait-runs", type=int, nargs='?', const=True, default=1, help="maximum idle runs to wait for incoming data")


# def str2bool_a(v: str) -> bool:
#     try:
#         return str2bool(v)
#     except ValueError:
#         raise argparse.ArgumentTypeError('Boolean value expected.')


def init_command_wrapper(pipeline_name: str, destination_name: str, branch: str) -> None:
    try:
        init_command(pipeline_name, destination_name, branch)
    except Exception as ex:
        click.secho(str(ex), err=True, fg="red")
        # TODO: display stack trace if with debug flag


def main() -> None:
    parser = argparse.ArgumentParser(description="Runs various DLT modules", formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    subparsers = parser.add_subparsers(dest="command")

    # dbt = subparsers.add_parser("dbt", help="Executes dbt package")
    # add_pool_cli_arguments(dbt)
    init_cmd = subparsers.add_parser("init", help="Creates new pipeline script from a selected template.")
    init_cmd.add_argument("pipeline_name", help="Pipeline name. If pipeline with given name already exists it will be used as a template. Otherwise new template will be created.")
    init_cmd.add_argument("destination_name", help="Name of the destination ie. bigquery or redshift")
    init_cmd.add_argument("--branch", default=None, help="Advanced. Uses specific branch of the init repository to fetch the template.")

    schema = subparsers.add_parser("schema", help="Shows, converts and upgrades schemas")
    schema.add_argument("file", help="Schema file name, in yaml or json format, will autodetect based on extension")
    schema.add_argument("--format", choices=["json", "yaml"], default="yaml", help="Display schema in this format")
    schema.add_argument("--remove-defaults", action="store_true", help="Does not show default hint values")

    pipe_cmd = subparsers.add_parser("pipeline", help="Operations on the pipelines")
    pipe_cmd.add_argument("name", help="Pipeline name")
    pipe_cmd.add_argument("operation", choices=["failed_loads", "drop", "init"], default="failed_loads", help="Show failed loads for a pipeline")
    pipe_cmd.add_argument("--workdir", help="Pipeline working directory", default=None)

    args = parser.parse_args()
    # run_f: Callable[[TRunnerArgs], None] = None

    # if args.command == "normalize":
    #     from dlt.normalize.normalize import run_main as normalize_run
    #     run_f = normalize_run
    # elif args.command == "load":
    #     from dlt.load.load import run_main as loader_run
    #     run_f = loader_run
    # if args.command == "dbt":
    #     from dlt.dbt_runner.runner import run_main as dbt_run
    #     run_f = dbt_run
    if args.command == "schema":
        with open(args.file, "r", encoding="utf-8") as f:
            if os.path.splitext(args.file)[1][1:] == "json":
                schema_dict: DictStrAny = json.load(f)
            else:
                schema_dict = yaml.safe_load(f)
        s = Schema.from_dict(schema_dict)
        if args.format == "json":
            schema_str = json.dumps(s.to_dict(remove_defaults=args.remove_defaults), indent=2)
        else:
            schema_str = s.to_pretty_yaml(remove_defaults=args.remove_defaults)
        print(schema_str)
        exit(0)
    elif args.command == "pipeline":

        p = attach(pipeline_name=args.name, working_dir=args.workdir)
        print(f"Found pipeline {p.pipeline_name} ({args.name}) in {p.working_dir} ({args.workdir}) with state {p._get_state()}")

        if args.operation == "failed_loads":
            completed_loads = p.list_completed_load_packages()
            for load_id in completed_loads:
                print(f"Checking failed jobs in load id '{load_id}'")
                for job, failed_message in p.list_failed_jobs_in_package(load_id):
                    print(f"JOB: {os.path.abspath(job)}\nMSG: {failed_message}")

        if args.operation == "drop":
            p.drop()

        exit(0)
    elif args.command == "init":
        init_command_wrapper(args.pipeline_name, args.destination_name, args.branch)
        exit(0)
    else:
        parser.print_help()
        exit(-1)

    # run_args = TRunnerArgs(args.is_single_run, args.wait_runs)
    # run_f(run_args)

if __name__ == "__main__":
    main()




# from dlt.common import json
# from dlt.common.schema import Schema
# from dlt.common.typing import DictStrAny

# from dlt.pipeline import Pipeline, PostgresPipelineCredentials

# # this is example of extracting from iterator that may be a
# # - jsonl file you read line by line
# # - a list of dicts
# # - iterator over panda frame https://pandas.pydata.org/docs/reference/api/pandas.DataFrame.to_dict.html
# # - any iterator with map function
# # - any iterator whatsoever ie getting data from REST API

# # extracting from iterator always happens in single worker thread and iterator is directly returning data so there's no retry mechanism if anything fails

# # working BQ creds
# # credentials = Pipeline.load_gcp_credentials("_secrets/project1234_service.json", "gamma_guild")

# if __name__ == '__main__':
#     # working redshift creds, you can pass password as last parameter or via PG__PASSWORD env variable ie.
#     # LOG_LEVEL=INFO PG__PASSWORD=.... python examples/discord_iterator.py
#     credentials = PostgresPipelineCredentials("redshift", "chat_analytics_rasa", "gamma_guild_8", "loader", "3.73.90.3")

#     pipeline = Pipeline("discord")
#     schema: Schema = None
#     # uncomment to use already modified schema, but the auto-inferred schema will also work nicely
#     # TODO: schema below needs is not yet finished, more exploration needed
#     # schema = Pipeline.load_schema_from_file("examples/schemas/discord_schema.yml")
#     pipeline.create_pipeline(credentials, schema=schema)
#     # the pipeline created a working directory. you can always attach the pipeline back by just providing the dir to
#     # Pipeline::restore_pipeline
#     print(pipeline.root_path)

#     # load our sample data - list of discord channels
#     with open("examples/data/channels.json", "r", encoding="utf-8") as f:
#         channels = json.load(f)
#     # and extract it
#     pipeline.extract(channels, table_name="channels")
#     # load list of messages
#     with open("examples/data/messages.json", "r", encoding="utf-8") as f:
#         messages = json.load(f)

#     # but we actually want to process messages before processing so pass mapping function instead
#     def processor(m: DictStrAny) -> DictStrAny:
#         if "referenced_message" in m and m["referenced_message"] is not None:
#             # we do not want full entity for referenced message, just the id
#             m["referenced_message"] = m["referenced_message"]["id"]

#         return m

#     # pass mapping iterator over messages and extract
#     pipeline.extract(map(processor, messages), table_name="messages")

#     # from now on each pipeline does more or less the same thing: normalize and load data

#     # now create loading packages and infer the schema
#     pipeline.normalize()

#     # show loads, each load contains a copy of the schema that corresponds to the data inside
#     # and a set of directories for job states (new -> in progress -> failed|completed)
#     new_loads = pipeline.list_normalized_loads()

#     # get inferred schema
#     # schema = pipeline.get_current_schema()
#     # print(schema.as_yaml(remove_default=True))

#     # load packages
#     pipeline.load()

#     # should be empty
#     new_loads = pipeline.list_normalized_loads()
#     print(new_loads)

#     # now enumerate all complete loads if we have any failed packages
#     # complete but failed job will not raise any exceptions
#     completed_loads = pipeline.list_completed_loads()
#     # print(completed_loads)
#     for load_id in completed_loads:
#         print(f"Checking failed jobs in {load_id}")
#         for job, failed_message in pipeline.list_failed_jobs(load_id):
#             print(f"JOB: {job}\nMSG: {failed_message}")



from typing import Type
from dlt.common.data_writers.escape import escape_bigquery_identifier

from dlt.common.schema.schema import Schema
from dlt.common.configuration import with_config
from dlt.common.configuration.accessors import config
from dlt.common.destination import DestinationCapabilitiesContext, JobClientBase, DestinationClientConfiguration

from dlt.destinations.bigquery.configuration import BigQueryClientConfiguration


@with_config(spec=BigQueryClientConfiguration, namespaces=("destination", "bigquery",))
def _configure(config: BigQueryClientConfiguration = config.value) -> BigQueryClientConfiguration:
    return config


def capabilities() -> DestinationCapabilitiesContext:
    caps = DestinationCapabilitiesContext()
    caps.update({
        "preferred_loader_file_format": "jsonl",
        "supported_loader_file_formats": ["jsonl"],
        "escape_identifier": escape_bigquery_identifier,
        "escape_literal": None,
        "max_identifier_length": 1024,
        "max_column_identifier_length": 300,
        "max_query_length": 1024 * 1024,
        "is_max_query_length_in_bytes": False,
        "max_text_data_type_length": 10 * 1024 * 1024,
        "is_max_text_data_type_length_in_bytes": True
    })
    return caps


def client(schema: Schema, initial_config: DestinationClientConfiguration = config.value) -> JobClientBase:
    # import client when creating instance so capabilities and config specs can be accessed without dependencies installed
    from dlt.destinations.bigquery.bigquery import BigQueryClient

    return BigQueryClient(schema, _configure(initial_config))  # type: ignore


def spec() -> Type[DestinationClientConfiguration]:
    return BigQueryClientConfiguration


import abc
import contextlib
import inspect
from typing import Any, ClassVar, Sequence, Type, TypeVar
from dlt.common import json
from dlt.common.configuration.container import Container

from dlt.common.configuration.providers.provider import ConfigProvider
from dlt.common.configuration.resolve import deserialize_value
from dlt.common.configuration.specs.base_configuration import is_base_configuration_hint
from dlt.common.configuration.specs.config_providers_context import ConfigProvidersContext
from dlt.common.schema.utils import coerce_value
from dlt.common.typing import AnyType, ConfigValue

DLT_SECRETS_VALUE = "secrets.value"
DLT_CONFIG_VALUE = "config.value"
TConfigAny = TypeVar("TConfigAny", bound=Any)

class _Accessor(abc.ABC):

    def __getitem__(self, field: str) -> Any:
        value = self._get_value(field)
        if value is None:
            raise KeyError(field)
        if isinstance(value, str):
            return self._auto_cast(value)
        else:
            return value


    def get(self, field: str, expected_type: Type[TConfigAny] = None) -> TConfigAny:
        value: TConfigAny = self._get_value(field, expected_type)
        if value is None:
            return None
        # cast to required type
        if expected_type:
            if is_base_configuration_hint(expected_type):
                c = expected_type()
                if isinstance(value, dict):
                    c.update(value)
                else:
                    c.parse_native_representation(value)
                return c  # type: ignore
            else:
                return deserialize_value(field, value, expected_type)  # type: ignore
        else:
            return value


    @property
    @abc.abstractmethod
    def config_providers(self) -> Sequence[ConfigProvider]:
        pass

    def _get_providers_from_context(self) -> Sequence[ConfigProvider]:
        return Container()[ConfigProvidersContext].providers

    def _auto_cast(self, value: str) -> Any:
        # try to cast to bool, int, float and complex (via JSON)
        if value.lower() == "true":
            return True
        if value.lower() == "false":
            return False
        with contextlib.suppress(ValueError):
            return coerce_value("bigint", "text", value)
        with contextlib.suppress(ValueError):
            return coerce_value("double", "text", value)
        with contextlib.suppress(ValueError):
            c_v = json.loads(value)
            # only lists and dictionaries count
            if isinstance(c_v, (list, dict)):
                return c_v
        return value

    def _get_value(self, field: str, type_hint: Type[Any] = AnyType) -> Any:
        # split field into namespaces and a key
        namespaces = field.split(".")
        key = namespaces.pop()
        value = None
        for provider in self.config_providers:
            value, _ = provider.get_value(key, type_hint, *namespaces)
            if value is not None:
                break
        return value


class _ConfigAccessor(_Accessor):
    """Provides direct access to configured values that are not secrets."""

    @property
    def config_providers(self) -> Sequence[ConfigProvider]:
        """Return a list of config providers, in lookup order"""
        return [p for p in self._get_providers_from_context()]

    value: ClassVar[None] = ConfigValue
    "A placeholder that tells dlt to replace it with actual config value during the call to a source or resource decorated function."


class _SecretsAccessor(_Accessor):
    """Provides direct access to secrets."""

    @property
    def config_providers(self) -> Sequence[ConfigProvider]:
        """Return a list of config providers that can hold secrets, in lookup order"""
        return [p for p in self._get_providers_from_context() if p.supports_secrets]

    value: ClassVar[None] = ConfigValue
    "A placeholder that tells dlt to replace it with actual secret during the call to a source or resource decorated function."


config = _ConfigAccessor()
"""Dictionary-like access to all secrets known to dlt"""

secrets = _SecretsAccessor()
"""Dictionary-like access to all config values known to dlt"""



import platform

from dlt.destinations.postgres.sql_client import Psycopg2SqlClient
if platform.python_implementation() == "PyPy":
    import psycopg2cffi as psycopg2
    # from psycopg2cffi.sql import SQL, Composed
else:
    import psycopg2
    # from psycopg2.sql import SQL, Composed

from typing import Dict, Optional

from dlt.common.arithmetics import DEFAULT_NUMERIC_PRECISION, DEFAULT_NUMERIC_SCALE
from dlt.common.destination import DestinationCapabilitiesContext
from dlt.common.schema import TColumnSchema, TDataType, TColumnHint, Schema

from dlt.destinations.exceptions import DatabaseTerminalException

from dlt.destinations.redshift import capabilities
from dlt.destinations.redshift.configuration import RedshiftClientConfiguration
from dlt.destinations.postgres.postgres import PostgresClientBase


SCT_TO_PGT: Dict[TDataType, str] = {
    "complex": "varchar(max)",
    "text": "varchar(max)",
    "double": "double precision",
    "bool": "boolean",
    "timestamp": "timestamp with time zone",
    "bigint": "bigint",
    "binary": "varbinary",
    "decimal": f"numeric({DEFAULT_NUMERIC_PRECISION},{DEFAULT_NUMERIC_SCALE})"
}

PGT_TO_SCT: Dict[str, TDataType] = {
    "varchar(max)": "text",
    "double precision": "double",
    "boolean": "bool",
    "timestamp with time zone": "timestamp",
    "bigint": "bigint",
    "binary varying": "binary",
    "numeric": "decimal"
}

HINT_TO_REDSHIFT_ATTR: Dict[TColumnHint, str] = {
    "cluster": "DISTKEY",
    # it is better to not enforce constraints in redshift
    # "primary_key": "PRIMARY KEY",
    "sort": "SORTKEY"
}


class RedshiftSqlClient(Psycopg2SqlClient):
    @staticmethod
    def _maybe_make_terminal_exception_from_data_error(pg_ex: psycopg2.DataError) -> Optional[Exception]:
        if "Cannot insert a NULL value into column" in pg_ex.pgerror:
            # NULL violations is internal error, probably a redshift thing
            return DatabaseTerminalException(pg_ex)
        if "Numeric data overflow" in pg_ex.pgerror:
            return DatabaseTerminalException(pg_ex)
        if "Precision exceeds maximum" in pg_ex.pgerror:
            return DatabaseTerminalException(pg_ex)
        return None


class RedshiftClient(PostgresClientBase):

    def __init__(self, schema: Schema, config: RedshiftClientConfiguration) -> None:
        sql_client = RedshiftSqlClient (
            schema.normalize_make_dataset_name(config.dataset_name, config.default_schema_name, schema.name),
            config.credentials
        )
        super().__init__(schema, config, sql_client)
        self.sql_client = sql_client
        self.config: RedshiftClientConfiguration = config
        self.caps = self.capabilities()

    def _get_column_def_sql(self, c: TColumnSchema) -> str:
        hints_str = " ".join(HINT_TO_REDSHIFT_ATTR.get(h, "") for h in HINT_TO_REDSHIFT_ATTR.keys() if c.get(h, False) is True)
        column_name = self.caps.escape_identifier(c["name"])
        return f"{column_name} {self._sc_t_to_pq_t(c['data_type'])} {hints_str} {self._gen_not_null(c['nullable'])}"

    @staticmethod
    def _sc_t_to_pq_t(sc_t: TDataType) -> str:
        if sc_t == "wei":
            return f"numeric({DEFAULT_NUMERIC_PRECISION},0)"
        return SCT_TO_PGT[sc_t]

    @staticmethod
    def _pq_t_to_sc_t(pq_t: str, precision: Optional[int], scale: Optional[int]) -> TDataType:
        if pq_t == "numeric":
            if precision == DEFAULT_NUMERIC_PRECISION and scale == 0:
                return "wei"
        return PGT_TO_SCT.get(pq_t, "text")

    @classmethod
    def capabilities(cls) -> DestinationCapabilitiesContext:
        return capabilities()



import re
import ast
import inspect
import astunparse
from typing import Dict, List, Optional

from dlt.common.typing import AnyFun


def get_literal_defaults(node: ast.FunctionDef) -> Dict[str, str]:
    defaults: List[ast.expr] = []
    if node.args.defaults:
        defaults.extend(node.args.defaults)
    if node.args.kw_defaults:
        defaults.extend(node.args.kw_defaults)
    args: List[ast.arg] = []
    if node.args.posonlyargs:
        args.extend(node.args.posonlyargs)
    if node.args.args:
        args.extend(node.args.args)
    if node.args.kwonlyargs:
        args.extend(node.args.kwonlyargs)

    # zip args and defaults
    literal_defaults: Dict[str, str] = {}
    for arg, default in zip(reversed(args), reversed(defaults)):
        if default:
            literal_defaults[str(arg.arg)] = astunparse.unparse(default).strip()

    return literal_defaults


def get_func_def_node(f: AnyFun) -> ast.FunctionDef:
    # this will be slow
    source, lineno = inspect.findsource(inspect.unwrap(f))

    for node in ast.walk(ast.parse("".join(source))):
        if isinstance(node, ast.FunctionDef):
            f_lineno = node.lineno - 1
            # get line number of first decorator
            if node.decorator_list:
                f_lineno = node.decorator_list[0].lineno - 1
            # line number and function name must match
            if f_lineno == lineno and node.name == f.__name__:
                return node
    return None


def get_outer_func_def(node: ast.AST) -> Optional[ast.FunctionDef]:
    if not hasattr(node, "parent"):
        raise ValueError("No parent information in node, not enabled in visitor", node)
    while not isinstance(node.parent, ast.FunctionDef):  # type: ignore
        if node.parent is None:  # type: ignore
            return None
        node = node.parent  # type: ignore
    return node  # type: ignore


def set_ast_parents(tree: ast.AST) -> None:
    for node in ast.walk(tree):
        for child in ast.iter_child_nodes(node):
            child.parent = node if node is not tree else None  # type: ignore
